#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import bisect
from contextlib import suppress
import datetime
import numpy as np
from pricelib.common.utilities.enums import StatusType, ExerciseType
from pricelib.common.time import (global_evaluation_date, Schedule, CN_CALENDAR, AnnualDays)
from pricelib.common.utilities.patterns import Observer
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.product_base.option_base import OptionBase
from pricelib.pricing_engines.fdm_engines import FdmSnowBallEngine, FdmPhoenixEngine


class AutocallableBase(OptionBase, Observer):
    """自动赎回结构定价基类：二元小雪球Autocall Note(无敲入)/雪球式Snowball Autocallable/触发器Trigger Autocallable
                        二元小雪球：无敲入（敲入价为-1），有红利票息（未敲出有保底收益），为保本产品
                        雪球式：有敲入，敲入后下跌部分保护，敲出年化票息
                        触发器：有敲入，敲入后下跌部分保护，敲出绝对票息
       支持变敲出、变票息、变敲入、雪球plus、保底雪球、OTM、触发式等多种结构"""

    def __init__(self, s0, barrier_out=None, barrier_in=None, coupon_out=None, coupon_div=None, parti_out=0,
                 parti_in=1, margin_lvl=1, strike_upper=None, strike_lower=0, strike_call=None, obs_dates=None,
                 pay_dates=None, lock_term=1, trigger=False, in_obs_type=ExerciseType.American,
                 status=StatusType.NoTouch, engine=None, maturity=None, start_date=None, end_date=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        """构造函数
        产品参数:
            s0: float，标的初始价格
            barrier_out: List[float]，敲出障碍价，绝对值/百分比，长度需要与敲出观察日数一致
                         若敲出价不是常数，为阶梯式雪球（降敲）
            barrier_in: List[float]，敲入障碍价，绝对值/百分比，仅雪球有，值大于0时非保本，值为0时保本
                        若敲入价不是常数，为变敲入型雪球，每个敲出观察日对应一个敲入价，在这个敲出观察日和上个敲出观察日之间，敲入价为该敲入价
            coupon_out: List[float]，敲出票息，百分比，年化
                        数组list，长度需要与敲出观察日数一致，为阶梯式浮动票息（降息或双降）雪球，主流降票息结构为早利(2年)和蝶变(3年)
            coupon_div: float，百分比，年化，雪球为红利票息；Autocall Note为未敲出票息
                        是持有到期未敲入未敲出的票息，与敲出票息分开，为数值型，不能是数组，默认为None，即与最后一个敲出票息相同
            parti_out: float，看涨结构参与率
            parti_in: float，敲入后参与率，限损
            margin_lvl: float，保证金比例，默认为1，即无杠杆
            strike_upper: float，敲入发生后(熊市价差)的高行权价，即OTM行权价(<s0)，普通雪球敲入发生后的高行权价为期初价格s0
            strike_lower: float，敲入发生后(熊市价差)的低行权价，即保底边界，普通雪球敲入发生后的低行权价为0
                                对于雪球和触发器，为保底价，越高提供的保护越多，值为0时不保底
            strike_call: float，敲出价之上的看涨结构行权价
            lock_term: int，锁定期，单位为月，锁定期内不触发敲出
            trigger: bool，是否为触发器，即敲出票息是否年化，True为绝对值，False为百分比
            in_obs_type: 敲入观察类型，ExerciseType枚举类，默认为American每日观察；European为仅到期日观察是否敲入
            status: 敲入敲出状态，StatusType枚举类，默认为NoTouch, 即未敲入未敲出；UpTouch为已敲出，DownTouch为已敲入
            engine: 定价引擎，PricingEngine类
                    蒙特卡洛: MCAutoCallableEngine
                    PDE: FdmSnowBallEngine
                    积分法: QuadSnowballEngine
        时间参数: 要么输入年化期限，要么输入起始日和到期日；敲出观察日和票息支付日可缺省
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            obs_dates: List[datetime.date]，敲出观察日，可缺省，缺省时会自动生成每月观察的敲出日期序列(已根据节假日调整)
            pay_dates: List[datetime.date]，票息支付日，可缺省，长度需要与敲出观察日数一致，如不指定则默认为敲出观察日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
        可选参数:
            若未提供引擎的情况下，提供了标的价格、无风险利率、分红/融券率、波动率，
            则默认使用PDE定价引擎 FdmSnowBallEngine
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.s0 = s0  # 标的初始价格
        self.trade_calendar = trade_calendar  # 交易日历
        self.annual_days = annual_days
        self.t_step_per_year = t_step_per_year
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        if maturity is None:
            self.maturity = (end_date - start_date).days / annual_days.value
        else:
            self.maturity = maturity
        self.barrier_out = barrier_out  # 敲出线
        self.barrier_in = barrier_in  # 敲入线
        self.coupon_out = coupon_out  # 敲出票息
        self.coupon_div = coupon_out if coupon_div is None else coupon_div  # 红利票息
        self.parti_out = parti_out  # 上涨参与率
        self.parti_in = parti_in  # 下跌参与率
        self.margin_lvl = margin_lvl  # 预付金比例
        self.strike_upper = s0 if strike_upper is None else strike_upper  # 敲入后高执行价
        self.strike_lower = strike_lower  # 低行权价
        self.strike_call = s0 if strike_call is None else strike_call  # 敲出看涨执行价
        if obs_dates is None:  # 观察日列表
            self.obs_dates = Schedule(trade_calendar=trade_calendar, start=self.start_date, end=self.end_date,
                                      freq='m', lock_term=lock_term)
            self.end_date = self.obs_dates.end  # 修正结束时间
        else:
            self.obs_dates = Schedule(trade_calendar=trade_calendar, date_schedule=obs_dates)
        if pay_dates is None:  # 付息日
            self.pay_dates = self.obs_dates
        else:
            self.pay_dates = Schedule(trade_calendar=trade_calendar, date_schedule=pay_dates)
        self.lock_term = lock_term  # 锁定期
        self.trigger = trigger  # 是否是触发式
        self.status = status
        self.in_obs_type = in_obs_type  # 敲入观察为欧式，到期观察敲入；美式，每日观察敲入
        if engine is not None:
            self.set_pricing_engine(engine)
        elif s is not None and r is not None and q is not None and vol is not None:
            default_engine = FdmSnowBallEngine(s=s, r=r, q=q, vol=vol)
            self.set_pricing_engine(default_engine)

    def set_pricing_engine(self, engine):
        """设置定价引擎，同时将自己注册为观察者。若已有定价引擎，先将自己从原定价引擎的观察者列表中移除"""
        with suppress(AttributeError, ValueError):
            self.engine.process.spot.remove_observer(self)
        self.engine = engine
        self.engine.process.spot.add_observer(self)
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.engine.process.spot.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """收到标的价格变化的通知时，自动更新自动赎回结构是否已经敲入或敲出"""
        if observable == self.engine.process.spot:
            calculate_date = global_evaluation_date()
            if calculate_date in self.obs_dates:  # 如果估值日是敲出观察日
                obs_idx = self.obs_dates.date_schedule.index(calculate_date)
                if observable.value >= self.barrier_out[obs_idx]:  # 如果标的价格大于等于敲出价，敲出
                    self.status = StatusType.UpTouch
                    return
            if self.status in [StatusType.DownTouch, StatusType.UpTouch]:  # 如果已敲入/已敲出，维持原来的状态
                return
            if self.status == StatusType.NoTouch:  # 如果未敲入未敲出
                obs_idx = bisect.bisect_left(self.obs_dates.date_schedule, global_evaluation_date())
                if observable.value <= self.barrier_in[obs_idx]:  # 且标的价格小于等于敲入价，敲入
                    self.status = StatusType.DownTouch
                    return

    def __repr__(self):
        """返回期权的描述"""
        return "Autocallable自动赎回结构"

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price


class PhoenixBase(OptionBase, Observer):
    """FCN/DCN/凤凰结构定价基类
        凤凰类 AutoCallable
            较雪球式AutoCallable，增加派息价格，少了敲出和红利票息
            每个派息（敲出）观察日，如果价格高于派息价格，派固定利息，如果发生敲出，合约提前结束；
            发生敲入后，派息方式不变，到期如果未敲出，结构为看跌空头或熊市价差空头
        FCN(Fixed Coupon Note, 固定派息票据)/DCN(Digital Coupon Note, 二元派息票据)/Phoenix凤凰票据
            FCN：无派息价格（派息价格为0），每月固定派息；到期日观察是否敲入
            DCN：有派息价格，每个观察日如果价格高于派息价格，派固定利息；到期日观察是否敲入
            Phoenix：有派息价格，每个观察日如果价格高于派息价格，派固定利息；每日观察是否敲入
    """

    def __init__(self, s0,
                 barrier_out=None, barrier_yield=None, coupon=None, barrier_in=None, parti_in=1, margin_lvl=1,
                 strike_upper=None, strike_lower=0, obs_dates=None, pay_dates=None, lock_term=1,
                 in_obs_type=ExerciseType.American, status=StatusType.NoTouch, engine=None, maturity=None,
                 start_date=None, end_date=None, trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365,
                 t_step_per_year=243, s=None, r=None, q=None, vol=None):
        """构造函数
        Args:
            s0: float，标的初始价格
            barrier_out: List[float]，敲出障碍价，绝对值/百分比，长度需要与敲出观察日数一致
                         若不是常数，为阶梯式凤凰（降敲）
            barrier_yield: List[float]，派息边界价格，绝对值/百分比，长度需要与派息观察日数一致
                           凤凰和DCN，barrier_yield一般等于敲入价格；FCN每月固定派息，barrier_yield等于0
            coupon: List[float]，派息票息，百分比，非年化
                    数组list，长度需要与派息观察日数一致，为阶梯式浮动票息（降息或双降）凤凰
            barrier_in: List[float]，敲入障碍价，绝对值/百分比
                    若不是常数，为变敲入型雪球，每个敲出观察日对应一个敲入价，在这个敲出观察日和上个敲出观察日之间，敲入价为该敲入价
                    对于FCN，进到期日观察是否敲入，到期日前的敲入价都是0
            parti_in: float，敲入后参与率，限损
            margin_lvl: float，保证金比例，默认为1，即无杠杆
            strike_upper: float，敲入发生后(熊市价差)的高行权价，即OTM行权价(<s0)，普通凤凰敲入发生后的高行权价为期初价格s0
            strike_lower: float，敲入发生后(熊市价差)的低行权价，即保底边界，越高提供的保护越多，值为0时不保底
            lock_term: int，锁定期，单位为月，锁定期内不触发敲出
            status: 敲入敲出状态，StatusType枚举类，默认为NoTouch，已敲入未敲出为DownTouch，已敲出为UpTouch
            in_obs_type: 敲入观察类型，ExerciseType枚举类，默认为American每日观察；European为仅到期日观察是否敲入
            engine: 定价引擎，PricingEngine类
                    蒙特卡洛: MCPhoenixEngine
                    PDE: FdmPhoenixEngine
                    积分法: QuadFCNEngine
        时间参数: 要么输入年化期限，要么输入起始日和到期日；敲出观察日和票息支付日可缺省
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            obs_dates: List[datetime.date]，敲出观察日，可缺省，缺省时会自动生成每月观察的敲出日期序列(已根据节假日调整)
            pay_dates: List[datetime.date]，票息支付日，可缺省，长度需要与敲出观察日数一致，如不指定则默认为敲出观察日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
        可选参数:
            若未提供引擎的情况下，提供了标的价格、无风险利率、分红/融券率、波动率，
            则默认使用PDE定价引擎 FdmSnowBallEngine
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.s0 = s0  # 标的初始价格
        self.trade_calendar = trade_calendar  # 交易日历
        self.annual_days = annual_days
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        if maturity is None:
            self.maturity = (end_date - start_date).days / annual_days.value
        else:
            self.maturity = maturity
        len_months = round(self.maturity * 12)
        # obs_dates无锁定期，因为迭代过程中paydates和barrier in的改变都发生在敲出日obs_dates，锁定期靠敲出价+inf实现
        if obs_dates is None and pay_dates is None:  # 观察日列表
            self.obs_dates = self.pay_dates = Schedule(trade_calendar=trade_calendar, start=self.start_date,
                                                       end=self.end_date, freq='m', lock_term=1)
            self.end_date = self.obs_dates.end  # 修正结束时间
        elif obs_dates is not None and pay_dates is not None:
            assert len(pay_dates) == len_months, "Error: 输入派息日数组长度不等于存续期月份数"
            assert len(obs_dates) == len_months - lock_term + 1, "Error: 输入敲出观察日数组长度不等于存续期月份数减去锁定期"
            # todo: 区分敲出观察日和付息观察日，以及敲出支付日和付息日
            self.obs_dates = Schedule(trade_calendar=trade_calendar,
                                      date_schedule=pay_dates[:lock_term - 1] + obs_dates)
            self.pay_dates = Schedule(trade_calendar=trade_calendar, date_schedule=pay_dates)
        else:
            raise ValueError("观察日和付息日必须同时输入或同时不输入")
        self.t_step_per_year = t_step_per_year

        self.lock_term = lock_term  # 锁定期
        # 敲出线
        if barrier_out is not None:
            assert isinstance(barrier_out, (list, np.ndarray)) and len(barrier_out) == len_months, \
                f"Error: 输入的敲出线必须是与存续月份数{len_months}等长的list列表或np.adarray数组"
            self.barrier_out = np.array(barrier_out).astype(float)
            self.barrier_out[:(lock_term - 1)] = np.inf  # 锁定期敲出价格 为+inf
        else:
            self.barrier_out = barrier_out
        self.barrier_in = barrier_in  # 敲入线
        self.barrier_yield = barrier_yield  # 派息边界价格
        self.coupon = coupon  # 票息
        if coupon is not None:
            assert isinstance(coupon, (list, np.ndarray)) and len(coupon) == len_months, \
                f"Error: 输入的票息必须是与存续月份数{len_months}等长的list列表或np.adarray数组"
        self.parti_in = parti_in  # 下跌参与率
        self.margin_lvl = margin_lvl  # 预付金比例
        self.strike_upper = s0 if strike_upper is None else strike_upper  # 敲入后高执行价
        self.strike_lower = strike_lower  # 低行权价
        self.status = status
        self.in_obs_type = in_obs_type  # 敲入观察为欧式，到期观察敲入；美式，每日观察敲入
        if engine is not None:
            self.set_pricing_engine(engine)
        elif s is not None and r is not None and q is not None and vol is not None:
            default_engine = FdmPhoenixEngine(s=s, r=r, q=q, vol=vol)
            self.set_pricing_engine(default_engine)

    def set_pricing_engine(self, engine):
        with suppress(AttributeError, ValueError):
            self.engine.process.spot.remove_observer(self)
        self.engine = engine
        self.engine.process.spot.add_observer(self)
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.engine.process.spot.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """收到标的价格变化的通知时，自动更新自动赎回结构是否已经敲入或敲出"""
        if observable == self.engine.process.spot:
            # TODO: 目前默认敲入敲出是常数，尚未支持敲入敲出线可变
            if isinstance(self.barrier_in, (int, float)) and observable.value <= self.barrier_in:
                self.status = StatusType.DownTouch
            elif isinstance(self.barrier_out, (int, float)) and observable.value >= self.barrier_out:
                self.status = StatusType.UpTouch

    def __repr__(self):
        """返回期权的描述"""
        return "FCN/DCN/Phoenix"

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import StatusType
from pricelib.common.time import CN_CALENDAR, AnnualDays
from .autocallable_base import AutocallableBase


class FlooredSnowball(AutocallableBase):
    """保底雪球(限损雪球，不追保雪球)产品类"""

    def __init__(self, s0, barrier_out, barrier_in, coupon_out, strike_lower=None, margin_lvl=1,
                 coupon_div=None, lock_term=3, annual_days=AnnualDays.N365, parti_in=1,
                 maturity=None, start_date=None, end_date=None, trade_calendar=CN_CALENDAR, obs_dates=None,
                 pay_dates=None, engine=None, status=StatusType.NoTouch, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        """继承自动赎回基类AutocallableBase的参数，详见AutocallableBase的__init__方法
        Args: 以下为区别于经典雪球的参数
            strike_lower: float，敲入发生后(熊市价差)的低行权价，即保底边界，越高提供的保护越多，值为0时不保底
            margin_lvl: float，保证金比例，默认为1，即无杠杆。
                                不追保雪球一般将保证金比例设为最大可能亏损，例如75敲入，保证金比例25%，即为4倍杠杆
            parti_in: float，敲入后参与率，限损
        """
        super().__init__(s0=s0, maturity=maturity, start_date=start_date, end_date=end_date, lock_term=lock_term,
                         trade_calendar=trade_calendar, obs_dates=obs_dates, pay_dates=pay_dates, status=status,
                         annual_days=annual_days, parti_in=parti_in, margin_lvl=margin_lvl,
                         t_step_per_year=t_step_per_year, engine=engine, s=s, r=r, q=q, vol=vol)
        len_obs_dates = len(self.obs_dates.date_schedule)

        self.barrier_out = np.ones(len_obs_dates) * barrier_out
        self.barrier_in = np.ones(len_obs_dates) * barrier_in
        self.coupon_out = np.ones(len_obs_dates) * coupon_out
        self.coupon_div = coupon_div if coupon_div is not None else self.coupon_out[-1]
        self.parti_out = 0
        self.strike_upper = s0
        self.strike_lower = strike_lower if strike_lower is not None else barrier_in
        self.t_step_per_year = t_step_per_year

    def __repr__(self):
        """返回期权的描述"""
        return "保底雪球(限损雪球，不追保雪球)"



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.product_base.option_base import OptionBase


class UnderlyingAsset(OptionBase):
    """标的资产"""

    def __init__(self, s0, *, engine=None):
        super().__init__()
        self.s0 = s0
        if engine is not None:
            self.set_pricing_engine(engine)

    def set_pricing_engine(self, engine):
        self.engine = engine
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def __repr__(self):
        """返回期权的描述"""
        return f"期初价格为{self.s0}的标的资产"

    @time_this
    def price(self, t=None, spot=None):
        if spot is not None:
            return spot
        return self.engine.process.spot()

    def delta(self, *args, **kwargs):
        delta = 1
        return delta

    def gamma(self, *args, **kwargs):
        gamma = 0
        return gamma

    def vega(self, *args, **kwargs):
        vega = 0
        return vega

    def theta(self, *args, **kwargs):
        theta = 0
        return theta

    def rho(self, *args, **kwargs):
        rho = 0
        return rho



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import StatusType, ExerciseType
from pricelib.common.time import CN_CALENDAR, AnnualDays
from .autocallable_base import PhoenixBase


class FCN(PhoenixBase):
    """FCN(Fixed Coupon Note固定派息票据)产品类
    无派息价格（派息价格为0），每月固定派息；到期日观察是否敲入"""
    in_obs_type = ExerciseType.European

    def __init__(self, s0, barrier_out, barrier_in, coupon, obs_dates=None, pay_dates=None, strike_upper=None,
                 lock_term=1, parti_in=1, margin_lvl=1, status=StatusType.NoTouch,
                 engine=None, maturity=None, start_date=None, end_date=None, trade_calendar=CN_CALENDAR,
                 annual_days=AnnualDays.N365, t_step_per_year=243, s=None, r=None, q=None, vol=None):
        """
        Args:
            s0: float，标的初始价格
            barrier_out: float，敲出障碍价，绝对值/百分比
            barrier_in: float，敲入障碍价，绝对值/百分比
            coupon: float，派息票息，百分比，非年化
            strike_upper: float，敲入发生(熊市价差)的高行权价，即FCN敲入发生后的折价减仓价，一般为敲入价
            lock_term: int，锁定期，单位为月，锁定期内不触发敲出
            parti_in: float，敲入后参与率，限损
            margin_lvl: float，保证金比例，默认为1，即无杠杆
            status: 敲入敲出状态，StatusType枚举类，默认为NoTouch未敲入未敲出，UpTouch为已敲出
            engine: 定价引擎，PricingEngine类
                    蒙特卡洛: MCPhoenixEngine
                    PDE: FdmPhoenixEngine
                    积分法: QuadFCNEngine
        时间参数: 要么输入年化期限，要么输入起始日和到期日；敲出观察日和票息支付日可缺省
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            obs_dates: List[datetime.date]，敲出观察日，可缺省，缺省时会自动生成每月观察的敲出日期序列(已根据节假日调整)
            pay_dates: List[datetime.date]，票息支付日，可缺省，长度需要与敲出观察日数一致，如不指定则默认为敲出观察日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
        可选参数:
            若未提供引擎的情况下，提供了标的价格、无风险利率、分红/融券率、波动率，
            则默认使用 PDE 定价引擎 FdmPhoenixEngine
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(s0=s0, maturity=maturity, start_date=start_date, end_date=end_date, lock_term=lock_term,
                         trade_calendar=trade_calendar, obs_dates=obs_dates, pay_dates=pay_dates, status=status,
                         annual_days=annual_days, parti_in=parti_in, margin_lvl=margin_lvl,
                         t_step_per_year=t_step_per_year, engine=engine, s=s, r=r, q=q, vol=vol)
        len_pay_dates = len(self.pay_dates.date_schedule)
        # todo: 区分敲出观察日和付息观察日，以及敲出支付日和付息日
        # obs_dates无锁定期，因为迭代过程中paydates和barrier in的改变都发生在敲出日obs_dates，锁定期靠敲出价+inf实现

        self.lock_term = lock_term  # 锁定期
        # 锁定期敲出价格 为+inf
        self.barrier_out = np.ones(len_pay_dates) * barrier_out
        self.barrier_out[:(lock_term - 1)] = np.inf
        # FCN只有期末观察敲入
        self.barrier_in = np.zeros(len_pay_dates)
        self.barrier_in[-1] = barrier_in
        # FCN固定派息，派息线为0
        self.barrier_yield = np.zeros(len_pay_dates)
        self.coupon = np.ones(len_pay_dates) * coupon
        self.strike_upper = barrier_in if strike_upper is None else strike_upper
        self.strike_lower = 0
        self.t_step_per_year = t_step_per_year

    def __repr__(self):
        """返回期权的描述"""
        return "FCN(Fixed Coupon Note固定派息票据)"



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import CallPut
from pricelib.common.pricing_engine_base import McEngine
from pricelib.common.time import global_evaluation_date


class MCAccumulatorEngine(McEngine):
    """累购累沽 Monte Carlo 模拟定价引擎
    仅支持起初估值，不支持持仓估值，因为没有记录历史累积次数"""

    # pylint: disable=too-many-locals
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)
        _obs_dates = np.array(prod.obs_dates.count_business_days(calculate_date))
        if spot is None:
            spot = self.process.spot()
        else:
            self.reset_paths_flag()  # 重置路径标志位，重新生成路径

        paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                    t_step_per_year=prod.t_step_per_year).copy()

        # 敲出与行权日期
        barrier_out = np.ones(_obs_dates.size) * prod.barrier_out
        barrier_out = np.tile(barrier_out, (self.n_path, 1)).T
        strike = np.ones(_obs_dates.size) * prod.strike
        strike = np.tile(strike, (self.n_path, 1)).T
        if prod.callput == CallPut.Call:  # 累购
            knock_out_time_idx = np.argmax(paths[_obs_dates, :] >= barrier_out, axis=0)
            knock_in_time_idx = paths[_obs_dates, :] < strike
        else:  # 累沽
            knock_out_time_idx = np.argmax(paths[_obs_dates, :] <= barrier_out, axis=0)
            knock_in_time_idx = paths[_obs_dates, :] > strike

        knock_out_scenario = np.where(knock_out_time_idx > 0, True, False)
        knock_out_time_idx[~knock_out_scenario] = _obs_dates.size - 1
        knock_in_bool = np.where(np.arange(knock_in_time_idx.shape[0])[:, np.newaxis] > knock_out_time_idx, False,
                                 knock_in_time_idx)

        # 贴现因子
        discount_factor = np.empty(_obs_dates.size)
        for i, j in enumerate(_obs_dates):
            discount_factor[i] = self.process.interest.disc_factor(j / prod.t_step_per_year)
        n_underlying = np.sum(knock_in_bool, axis=0) * (prod.leverage_ratio - 1) + knock_out_time_idx
        exercise_price = paths[_obs_dates[knock_out_time_idx], np.arange(self.n_path)]

        payoff = n_underlying * (exercise_price - prod.strike) * discount_factor[knock_out_time_idx]
        return np.mean(payoff) * prod.callput.value



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import StatusType
from pricelib.common.time import CN_CALENDAR
from .autocallable_base import AutocallableBase


class ParachuteSnowball(AutocallableBase):
    """降落伞雪球产品类"""

    def __init__(self, s0, barrier_out, barrier_out_final, barrier_in, coupon_out, coupon_div=None,
                 lock_term=3, maturity=None, start_date=None, end_date=None, trade_calendar=CN_CALENDAR, obs_dates=None,
                 pay_dates=None, engine=None, status=StatusType.NoTouch, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        """继承自动赎回基类AutocallableBase的参数，详见AutocallableBase的__init__方法
        Args: 以下为区别于经典雪球的参数
            barrier_out: float，敲出障碍价，绝对值/百分比
            barrier_out_final: float，最后一个敲出观察日的敲出障碍价，绝对值/百分比
            coupon_div: float，百分比，年化，红利票息，即持有到期未敲入未敲出的票息
        """
        super().__init__(s0=s0, maturity=maturity, start_date=start_date, end_date=end_date, lock_term=lock_term,
                         trade_calendar=trade_calendar, obs_dates=obs_dates, pay_dates=pay_dates, status=status,
                         t_step_per_year=t_step_per_year, engine=engine, s=s, r=r, q=q, vol=vol)
        len_obs_dates = len(self.obs_dates.date_schedule)

        barrier_out = np.ones(len_obs_dates) * barrier_out
        barrier_out[-1] = barrier_out_final
        self.barrier_out = barrier_out
        self.barrier_in = np.ones(len_obs_dates) * barrier_in
        self.coupon_out = np.ones(len_obs_dates) * coupon_out
        self.coupon_div = coupon_div if coupon_div is not None else self.coupon_out[-1]
        self.parti_out = 0
        self.parti_in = 1
        self.margin_lvl = 1
        self.strike_upper = s0
        self.strike_lower = 0

    def __repr__(self):
        """返回期权的描述"""
        return "降落伞雪球"



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import importlib_metadata
from .pricing_engines import *
from .products import *
from .common import *
from .common.utilities import SimpleQuote

__author__ = '上海凌瓴信息科技有限公司: 马瑞祥, 夏鸿翔, 贡献: 张鹏任, 张峻尉'
__email__ = 'marx@galatech.com.cn'

try:
    __version__ = importlib_metadata.version("pricelib")
except importlib_metadata.PackageNotFoundError:
    __version__ = "dev"



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
import numpy as np
from pricelib.common.utilities.enums import CallPut, ExerciseType, PaymentType
from pricelib.common.pricing_engine_base import QuadEngine
from pricelib.common.time import global_evaluation_date


class QuadDigitalEngine(QuadEngine):
    """二元(数字)期权-数值积分法定价引擎"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        tau = prod.trade_calendar.business_days_between(calculate_date, prod.end_date) / prod.t_step_per_year
        if spot is None:
            spot = self.process.spot()

        r = self.process.interest(tau)
        q = self.process.div(tau)
        vol = self.process.vol(tau, spot)
        dt = prod.discrete_obs_interval
        self.backward_steps = round(tau / dt)
        self._check_method_params()

        # 设置积分法engine参数
        self.set_quad_params(r=r, q=q, vol=vol)

        # 初始化fft的对数价格向量及边界的对数价格向量
        self.init_grid(spot, vol, tau)
        # 行权价在价格格点上的对应格点
        strike_point = np.where(np.exp(self.ln_s_vec) >= prod.strike)[0][0]
        v_vec = np.zeros(len(self.ln_s_vec))
        if prod.callput == CallPut.Call:
            v_vec[strike_point:] = prod.rebate
        elif prod.callput == CallPut.Put:
            v_vec[:strike_point] = prod.rebate
        # 欧式：
        if prod.exercise_type == ExerciseType.European:
            value = self.fft_step_backward(np.log(np.array([spot])), self.ln_s_vec, v_vec, tau)[0]
        # 美式：
        elif prod.exercise_type == ExerciseType.American:
            for step in range(self.backward_steps - 1, 0, -1):
                if prod.payment_type == PaymentType.Hit:
                    rebate_v = prod.rebate
                elif prod.payment_type == PaymentType.Expire:
                    rebate_v = prod.rebate * math.exp(-r * dt * (self.backward_steps - step))
                else:
                    raise ValueError("Invalid payment type")
                if prod.callput == CallPut.Call:
                    v_vec[:strike_point] = self.fft_step_backward(self.ln_s_vec[:strike_point], self.ln_s_vec, v_vec, dt)
                    v_vec[strike_point:] = rebate_v
                elif prod.callput == CallPut.Put:
                    v_vec[strike_point:] = self.fft_step_backward(self.ln_s_vec[strike_point:], self.ln_s_vec, v_vec, dt)
                    v_vec[:strike_point] = rebate_v
                else:
                    raise ValueError("Invalid call put type")

            value = self.fft_step_backward(np.array(np.log([spot])), self.ln_s_vec, v_vec, dt)[0]
        else:
            raise ValueError("ExerciseType must be American or European")
        return value



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from abc import ABCMeta
from functools import lru_cache
import numpy as np
from scipy import sparse
from scipy.interpolate import interp1d
from ..processes import StochProcessBase
from ..utilities.numerical import LinearFlat, CubicSplineFlat, TDMA_ldu_jit
from ..utilities.patterns import HashableArray
from ..utilities.enums import ProcessType, EngineType
from .engine_base import PricingEngineBase


# pylint: disable=invalid-name
class FdmEngine(PricingEngineBase, metaclass=ABCMeta):
    """障碍期权PDE有限差分法定价引擎，适用于连续观察障碍期权价值"""
    engine_type = EngineType.PdeEngine

    def __init__(self, stoch_process: StochProcessBase = None, s_step=800, n_smax=2, fdm_theta=1, *,
                 s=None, r=None, q=None, vol=None):
        """初始化有限差分法定价引擎
        Args:
            stoch_process: StochProcessBase随机过程
            s_step: int，价格步数
            n_smax: int，价格网格上界，设为n_smax倍初始价格
            fdm_theta: float，时间方向有限差分theta，0：explicit, 1: implicit, 0.5: Crank-Nicolson
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process=stoch_process, s=s, r=r, q=q, vol=vol)
        self.s_step = max(s_step, 50)  # 价格步数
        self.n_smax = n_smax  # 价格网格上界s0倍数, 设为n_smax倍初始价格
        self.fdm_theta = fdm_theta  # 时间方向有限差分theta，0：explicit, 1: implicit, 0.5: Crank-Nicolson
        # 以下为计算过程的中间变量
        self.fdm = None  # 有限差分网格FdmGrid对象
        self.smax = None  # 价格网格上界

    def set_stoch_process(self, stoch_process: StochProcessBase):
        assert stoch_process.process_type == ProcessType.BSProcess1D, 'Error: PDE有限差分法目前只能使用1维BSM动态过程'
        self.process = stoch_process

    def delta(self, prod, t=0, spot=None, step=None):
        """求t时刻，价格spot处的delta值
        Args:
            prod: Product，产品对象
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
            step: float，微扰步长，默认为价格spot的1%
        Returns: delta = ∂V/∂S
        """
        spot = self.process.spot() if spot is None else spot
        step = spot * 0.01 if step is None else step
        f = self.fdm.functionize(self.fdm.v_grid[1:-1, t], kind='cubic')
        delta = (f(min(spot + step, self.smax)) - f(min(spot - step, self.smax))) / (2 * step)
        return delta

    def gamma(self, prod, t=0, spot=None, step=None):
        """求t时刻，价格spot处的gamma值
        Args:
            prod: Product，产品对象
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
            step: float，微扰步长，默认为价格spot的1%
        Returns: gamma = ∂2V/∂S2
        """
        spot = self.process.spot() if spot is None else spot
        step = spot * 0.01 if step is None else step
        f = interp1d(self.fdm.s_vec, self.fdm.v_grid[1:-1, t], kind='cubic')
        gamma = (f(spot + step) - 2 * f(spot) + f(spot - step)) / (step ** 2)
        return gamma

    def theta(self, prod, t=0, spot=None, step=1):
        """计算一天的theta值
        Args:
            prod: Product，产品对象
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
            step: int，时间步长，默认为1天
        Returns: theta = ∂V/∂t
        """
        spot = self.process.spot() if spot is None else spot
        f = self.fdm.functionize(self.fdm.v_grid[1:-1, t], kind='cubic')
        g = self.fdm.functionize(self.fdm.v_grid[1:-1, t + step], kind='cubic')
        theta = (g(spot) - f(spot)) / (step * spot)  # ( step * self.dt)
        return theta

    def vega(self, prod, t=0, spot=None, step=0.01):
        """计算波动率上升1%的vega todo:目前只支持常数波动率
        Args:
            prod: Product，产品对象
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
            step: float，微扰步长，默认为1%
        Returns: vega = ∂V/∂σ
        """
        spot = self.process.spot() if spot is None else spot
        last_vol = self.process.vol.volval
        last_grid = self.fdm.v_grid.copy()
        fdm_grid0 = self.fdm
        f = fdm_grid0.functionize(fdm_grid0.v_grid[1:-1, t], kind='cubic')
        self.process.vol.volval = last_vol + step
        self.calc_present_value(prod=prod)
        fdm_grid1 = self.fdm
        g = fdm_grid1.functionize(fdm_grid1.v_grid[1:-1, t], kind='cubic')
        vega = (g(spot) - f(spot)) / spot
        self.process.vol.volval = last_vol
        self.fdm.v_grid = last_grid
        return vega

    def rho(self, prod, t=0, spot=None, step=0.01):
        """计算无风险利率上升1%的rho todo:目前只支持常数无风险利率
        Args:
            prod: Product，产品对象
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
            step: float，微扰步长，默认为1%
        Returns: rho = ∂V/∂r
        """
        spot = self.process.spot() if spot is None else spot
        last_r = self.process.interest.data
        last_grid = self.fdm.v_grid.copy()
        fdm_grid0 = self.fdm
        f = fdm_grid0.functionize(fdm_grid0.v_grid[1:-1, t], kind='cubic')
        self.process.interest.data = last_r + step
        self.calc_present_value(prod=prod)
        fdm_grid1 = self.fdm
        g = fdm_grid1.functionize(fdm_grid1.v_grid[1:-1, t], kind='cubic')
        rho = (g(spot) - f(spot)) / spot
        self.process.interest.data = last_r
        self.fdm.v_grid = last_grid
        return rho

    def pv_and_greeks(self, prod, t=0, spot=None):
        """当一次性计算pv和5个greeks时，可以调用此函数，减少计算量
        Args:
            prod: Product，产品对象
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
        Returns:
            Dict[str:float]: {'pv': pv, 'delta': delta, 'gamma': gamma, 'theta': theta, 'vega': vega, 'rho': rho}
        """
        spot = self.process.spot() if spot is None else spot
        s_step = spot * 0.01
        f = self.fdm.functionize(self.fdm.v_grid[1:-1, t], kind='cubic')

        pv = np.float64(f(spot))
        delta = (f(min(spot + s_step, self.smax)) - f(min(spot - s_step, self.smax))) / (2 * s_step)
        gamma = (f(spot + s_step) - 2 * f(spot) + f(spot - s_step)) / (s_step ** 2)

        g = self.fdm.functionize(self.fdm.v_grid[1:-1, t + 1], kind='cubic')
        theta = (g(spot) - f(spot)) / (1 * spot)

        last_v_grid = self.fdm.v_grid

        last_vol = self.process.vol.volval
        self.process.vol.volval = last_vol + 0.01
        self.calc_present_value(prod=prod)
        g1 = self.fdm.functionize(self.fdm.v_grid[1:-1, t], kind='cubic')
        vega = (g1(spot) - f(spot)) / spot
        self.process.vol.volval = last_vol

        last_r = self.process.interest.data
        self.process.interest.data = last_r + 0.01
        self.calc_present_value(prod=prod)
        g2 = self.fdm.functionize(self.fdm.v_grid[1:-1, t], kind='cubic')
        rho = (g2(spot) - f(spot)) / spot
        self.process.interest.data = last_r

        self.fdm.v_grid = last_v_grid

        return {'pv': pv, 'delta': delta, 'gamma': gamma, 'vega': vega, 'theta': theta, 'rho': rho}

    def value_matrix(self):
        """返回S-t矩阵"""
        return self.fdm.v_grid, self.fdm.s_vec

    def delta_matrix(self, step=0):
        """计算S-t矩阵中，每个t时刻所有s格点的delta
            step: 标的价步长，默认为0，即选择PDE有限差分网格的价格步长"""
        if step == 0:
            # 默认PDE模拟时的步长 step = fdm_grid.ds
            delta_s_t = (self.fdm.v_grid[2:] - self.fdm.v_grid[:-2]) / (2 * self.fdm.ds)
            return delta_s_t, self.fdm.s_vec  # 中心差分，价格向量掐头0去尾smax，正好是fdm_grid.s_vec
        else:
            # 重新定义步长，在模拟矩阵的基础上插值
            n_step = np.round(self.smax / step).astype('int')
            spots = np.arange(0, n_step, 1) * step
            # 下方n_step和spots用于避免计算远离敲入水平的场景，降低运算量
            # n_step = 2 * np.round((self.spot - self.barrier_in + 1) / step)
            # spots = np.arange(0, n_step, 1) * step + self.barrier_in - 1
            delta_s_t = np.zeros((n_step - 1, self.fdm.v_grid.shape[1]))
            # 分别在fdm_grid.s_vec价格向量首位添加0和smax
            x_vec = np.insert(self.fdm.s_vec, 0, 0)
            x_vec = np.append(x_vec, self.smax)
            for j in range(self.fdm.v_grid.shape[1]):
                f = interp1d(x_vec, self.fdm.v_grid[:, j], kind='cubic')
                for i, spot in enumerate(spots[1:]):
                    delta_s_t[i, j] = (f(spot + step) - f(spot - step)) / (step * 2)
            return delta_s_t, spots

    def gamma_matrix(self, step=0):
        """计算S-t矩阵中，每个t时刻所有s格点的gamma
            step: 标的价步长，默认为0，即选择PDE有限差分网格的价格步长"""
        if step == 0:
            # 默认PDE模拟时的步长 step = fdm_grid.ds
            gamma_s_t = np.diff(self.fdm.v_grid, n=2, axis=0) / self.fdm.ds ** 2
            return gamma_s_t, self.fdm.s_vec
        else:
            # 重新定义步长，在模拟矩阵的基础上插值
            n_step = np.round(self.smax / step).astype('int')
            step = self.smax / n_step
            spots = np.arange(0, n_step, 1) * step

            gamma_s_t = np.zeros((n_step - 2, self.fdm.v_grid.shape[1]))
            # 分别在fdm_grid.s_vec价格向量首位添加0和smax
            x_vec = np.insert(self.fdm.s_vec, 0, 0)
            x_vec = np.append(x_vec, self.smax)
            for j in range(self.fdm.v_grid.shape[1]):
                f = interp1d(x_vec, self.fdm.v_grid[:, j], kind='cubic')
                for i, spot in enumerate(spots[1:-2]):
                    gamma_s_t[i, j] = (f(spot + step) - 2 * f(spot) + f(spot - step)) / (step ** 2)
            return gamma_s_t, spots

    def theta_matrix(self, step=0):
        """计算S-t矩阵中，每个s价格所有t时刻的theta'
            step: 时间步长，默认为0，即选择PDE有限差分网格的时间步长"""
        if step == 0:
            # 默认PDE模拟时的步长
            spot = self.process.spot()
            theta_s_t = np.diff(self.fdm.v_grid, axis=1) / spot  # / self.dt
            x_vec = np.insert(self.fdm.s_vec, 0, 0)
            x_vec = np.append(x_vec, self.smax)
            return theta_s_t, x_vec
        else:  # todo: 时间步长与pde有限差分网格不同的theta
            raise NotImplementedError("todo: 时间步长与pde有限差分网格不同的theta.")


class FdmGrid:
    """有限差分网格基类，设置系数矩阵和边界条件，定义递推方法
    网格的边界是[0, smax]；演化方法是逐天evolve，便于处理敲入覆盖未敲入的值
    适用于离散观察障碍PDE有限差分法定价，这样求得的是离散观察障碍期权的价值"""

    def __init__(self, smax, maturity, t_step_per_year=243, s_step=400, fn_pde_coef=None, fdm_theta=1):
        """初始化有限差分网格
        注意stdev、fwd、fn_abc以及bound的量纲必须一致，要么都是绝对值，要么都是除以F标准化的相对值
        Args:
            smax: float，价格网格上界
            t_step_per_year: int，每年的时间步数
            s_step: int，价格步数
            fn_pde_coef: function，根据时间t和价格向量s_vec，返回PDE系数相关向量a,b,c。设pde有限差分的价格向量索引为i_vec，
                        a是需要乘以i_vec^2的系数，b是需要乘以i_vec的系数，c是不需要乘以i_vec的系数
                        对于BSM而言，a是 vol^2, b是r-q, c是r
            fdm_theta: float，有限差分时间方向的theta方法，theta=0.5是Crank-Nicolson方法，theta=0是显式Euler方法，theta=1是隐式Euler方法
        """
        self._theta = fdm_theta
        self.t_step_per_year = t_step_per_year  # 每年的时间步数
        self.s_min = 0  # 价格向量下界
        self.s_max = smax  # 价格向量上界
        self.s_vec = np.linspace(self.s_min, self.s_max, int(max(s_step, 50)) + 1)[1:-1]  # s_vec 标的价格向量
        self.ds = (self.s_vec[-1] - self.s_vec[0]) / (self.s_vec.size - 1)  # 价格步长
        num = round(maturity * self.t_step_per_year)  # 时间格点数
        self.tv = np.linspace(0, maturity, num + 1)  # 时间向量
        self.dt = (self.tv[-1] - self.tv[0]) / (self.tv.size - 1)  # 时间步长
        self.i_vec = np.round(self.s_vec / self.ds).astype(int)  # s_vec 对应的索引向量 (1,2,...,s_step) 不含边界0和s_step
        self.fn_pde_coef = fn_pde_coef  # PDE系数函数，如果是倒向PDE，a是BSM的(sigma*S)^2, b是BSM的(r-q)*S,c是r
        self.eye = sparse.eye(self.s_vec.size)  # 单位矩阵
        self.v_grid = np.zeros((self.s_vec.size + 2, self.tv.size))  # 初始化一个零矩阵，用于记录每个时间t对应的期权价值向量
        # 以下为计算过程的中间变量
        self.lower = None  # 下对角线
        self.diag = None  # 主对角线
        self.upper = None  # 上对角线
        self.A = None
        self.M1 = None
        self.M2 = None

    @lru_cache(maxsize=1)  # 缓存函数结果，避免重复计算。(常数波动率、无风险利率、分红率时，每个时间点的系数矩阵都是一样的)
    def _set_matrix(self, a, b, c, dt):
        """设置线性方程组的系数矩阵，t是时间，yv是上一时间点的期权价值向量
        M1 yv{j} = M2 yv{j-1} - g{j-1} - g{j}，其中g{j}和g{j-1}是边界条件。注意如果是从后向前递推，j的时间比j-1的时间靠前
        fn_pde_coef返回预计算系数，a是需要乘以i_vec^2的系数，b是需要乘以i_vec的系数，c是不需要乘以i_vec的系数
        Args:
            a: np.ndarray，需要乘以i_vec^2的系数
            b: np.ndarray，需要乘以i_vec的系数
            c: np.ndarray，不需要乘以i_vec的系数
            dt: float，格点的时间步长
        Returns: None
        """
        # [1:-1]去掉最上面和最下面的边界条件点，因为这两个点的边界条件是已知的，不需要计算
        diffusion_square = a * self.i_vec ** 2  # 对于BSM而言，a是 vol^2
        drift = b * self.i_vec  # 对于BSM而言，b是 r-q
        #  系数矩阵预计算, l、d、u分别是对角线下方lower、主对角线diagonal和对角线上方upper的系数数组
        self.lower = 0.5 * (diffusion_square - drift)
        self.diag = -diffusion_square - c
        self.upper = 0.5 * (diffusion_square + drift)
        # 创建三对角矩阵A，用来构造半隐式矩阵M1，M2。M1是隐式部分，M2是显式部分
        self.A = sparse.diags((self.lower[1:], self.diag, self.upper[:-1]), (-1, 0, 1), format='csc') * dt  # csc格式的稀疏矩阵
        # j 时点的期权价值向量（未知量）的系数矩阵M1
        self.M1 = self.eye - self._theta * self.A
        # j - 1 时点的期权价值向量（已知量）的系数矩阵M2（对于隐式imp，M2是单位阵）
        self.M2 = self.eye + (1 - self._theta) * self.A

    def _set_vector(self, j, yv, dt):
        """设置矩阵方程等号右侧的 V
        V = 矩阵方程右侧，系数M2乘以j - 1 时点的期权价值向量（已知量）
        Args:
            j: int，第j个时间点(j的顺序是从后向前)
            yv: np.ndarray，j - 1 时点的期权价值向量（已知量）
            dt: float，格点的时间步长
        Returns: V，矩阵方程右侧，系数M2乘以j - 1 时点的期权价值向量（已知量）
        """
        v_vec = self.M2.dot(yv)
        # V的最上面和最下面，分别减去矩阵第一行的lower*yv[0]和最后一行的upper*yv[-1]，这两个值是已知的边界条件，不需要计算
        v_vec[0] += (self._theta * self.v_grid[0, j]
                     + (1 - self._theta) * self.v_grid[0, j + 1]) * self.lower[0] * dt
        v_vec[-1] += (self._theta * self.v_grid[-1, j]
                      + (1 - self._theta) * self.v_grid[-1, j + 1]) * self.upper[-1] * dt
        return v_vec  # 生成三对角矩阵及对应的向量

    def evolve(self, j, yv, dt):
        """根据时间 t 和 j - 1 时间点的期权价值向量yv，返回当前时间点的期权价值向量
        Args:
            j: int，第j个时间点(j的顺序是从后向前)
            yv: np.ndarray，j - 1 时间点的期权价值向量yv
            dt: float，格点的时间步长
        Returns: 当前时间点的期权价值向量
        """
        a, b, c = self.fn_pde_coef(self.tv[j], self.s_vec)
        a = HashableArray(a)
        # yv = self._step_condition(j, yv)  # 步骤条件
        self._set_matrix(a, b, c, dt)  # 设置线性方程组的系数矩阵
        v_vec = self._set_vector(j, yv, dt)  # 设置线性方程组的向量
        # yv = TDMA(self.M1, v_vec)  # 三对角矩阵Thomas算法求解
        # 三对角矩阵Thomas算法求解,jit加速
        yv = TDMA_ldu_jit(self.M1.diagonal(-1), self.M1.diagonal(0), self.M1.diagonal(1), v_vec)
        return yv

    def functionize(self, yv, kind="linear"):
        """返回插值函数，插值的x是价格向量self.s_vec，y是期权价值向量yv
        Args:
            yv: np.ndarray，期权价值向量
            kind: str，插值方法，linear线性插值/cubic三次样条插值
        Returns:
        """
        if kind == "linear":
            return LinearFlat(self.s_vec, yv)
        else:
            return CubicSplineFlat(self.s_vec, yv)


class FdmGridwithBound(FdmGrid):
    """连续观察障碍PDE有限差分网格对象
    与离散观察障碍网格的区别在于，连续观察障碍网格的上下界是障碍价格，而不是[0, smax]；evolve方法可以自行指定起止时间，支持观察时间窗口
    这样求得的期权价值是连续观察障碍期权的价值"""

    def __init__(self, smax, t_step_per_year=243, s_step=400,
                 fn_pde_coef=None, bound=(-np.inf, np.inf), fdm_theta=1):
        """初始化有限差分网格
        注意stdev、fwd、fn_abc以及bound的量纲必须一致，要么都是绝对值，要么都是除以F标准化的相对值
        Args:
            smax: float，价格网格上界
            t_step_per_year: int，每年的时间步数
            s_step: int，价格步数
            fn_pde_coef: function，根据时间t和价格向量s_vec，返回PDE系数相关向量a,b,c。设pde有限差分的价格向量索引为i_vec，
                        a是需要乘以i_vec^2的系数，b是需要乘以i_vec的系数，c是不需要乘以i_vec的系数
                        对于BSM而言，a是 vol^2, b是r-q, c是r
            fdm_theta: float，有限差分时间方向的theta方法，theta=0.5是Crank-Nicolson方法，theta=0是显式Euler方法，theta=1是隐式Euler方法
        """
        self._theta = fdm_theta
        self.t_step_per_year = t_step_per_year  # 每年的时间步数
        self.bound = bound  # 长度为2的元组，高低障碍价格
        self.s_min = max(bound[0], 0)  # 价格向量下界
        self.s_max = min(bound[1], smax)  # 价格向量上界
        # [1:-1]去掉最上面和最下面的边界条件点，因为这两个点的边界条件是已知的，不需要计算
        self.s_vec = np.linspace(self.s_min, self.s_max, int(max(s_step, 50)) + 1)[1:-1]  # s_vec 标的价格向量
        self.ds = (self.s_vec[-1] - self.s_vec[0]) / (self.s_vec.size - 1)  # 价格步长
        # self.i_vec = np.linspace(1, s_step - 1, self.s_vec.size)
        self.i_vec = np.round(self.s_vec / self.ds).astype(int)  # s_vec 对应的索引向量 (1,2,...,s_step) 不含边界0和s_step
        self.fn_pde_coef = fn_pde_coef  # PDE系数函数，如果是倒向PDE，a是BSM的(sigma*S)^2, b是BSM的(r-q)*S,c是r
        self.eye = sparse.eye(self.s_vec.size)  # 单位矩阵
        # 中间变量
        self.fn_bound = None  # 边界条件函数

    def set_boundary_condition(self, fn_bound):
        """设置边界条件
        Args:
            fn_bound: List[lambda]，长度为2的列表，元素为上下两个边界条件关于时间t的函数，
                      fn_bound[0]是下边界，fn_bound[1]是上边界。
        """
        self.fn_bound = fn_bound

    def _set_vector(self, j, yv, dt):
        """设置矩阵方程等号右侧的 V
        V = 矩阵方程右侧，系数M2乘以j - 1 时点的期权价值向量（已知量）
        Args:
            j: int，第j个时间点(j的顺序是从后向前)
            yv: np.ndarray，j - 1 时点的期权价值向量（已知量）
            dt: float，格点的时间步长
        Returns: V，矩阵方程右侧，系数M2乘以j - 1 时点的期权价值向量（已知量）
        """
        V = self.M2.dot(yv)
        # V的最上面和最下面，分别减去矩阵第一行的lower*yv[0]和最后一行的upper*yv[-1]，这两个值是已知的边界条件，不需要计算
        V[0] += (self._theta * self.fn_bound[0](self.tv[j])
                 + (1 - self._theta) * self.fn_bound[0](self.tv[j - 1])) * self.lower[0] * dt
        V[-1] += (self._theta * self.fn_bound[1](self.tv[j])
                  + (1 - self._theta) * self.fn_bound[1](self.tv[j - 1])) * self.upper[-1] * dt
        return V  # 生成三对角矩阵及对应的向量

    def evolve_with_interval(self, start, end, yv, x=None):
        """从start时间点递推到end时间点，更新yv
        若dt>0, 从前向后递推，local_vol校准; dt<0, 从后向前递推，有限差分法定价
        Args:
            start: float，起始时间点，距离估值日的年化期限
            end: float，结束时间点，距离估值日的年化期限
            yv: np.ndarray，start时刻的期权价值向量
            x: 若x不是None，将每个时间t对应的指定的x对应的y值记录下来。
               一般在计算敲入时，传入x=bound即障碍价格，用已敲入的值作为未敲入的边界条件
        Returns:
            yv: np.ndarray，end时刻的期权价值向量
            bc_bound: 边界条件函数，input时间t，output边界条件函数值
        """
        num = max(round(abs(start - end) * self.t_step_per_year), 1)  # 时间格点, 最小取1，是为了应对start=end的情况
        self.tv = np.linspace(start, end, num + 1)  # 生成时间格点
        self.dt = (self.tv[-1] - self.tv[0]) / (self.tv.size - 1)  # dt既可以是正数，也可以是负数
        # self.v_grid = np.zeros((self.s_vec.size + 2, self.tv.size))  # 初始化一个零矩阵，用于记录每个时间t对应的期权价值向量

        if x is None:
            for j in range(1, len(self.tv)):  # 从tv[1:]开始递推，因为tv[0]是已知的
                yv = self.evolve(j, yv, -self.dt)
                # self.v_grid[1:-1, j] = yv  # 记录每个时间t对应的期权价值向量
            return yv
        else:  # 如果需要将每个时间t对应的指定的x对应的y值记录下来return，则需传入x
            # 一般用于计算敲入时，传入x=bound即障碍价格，用已敲入的期权价值作为未敲入的边界条件
            fv = np.empty((self.tv.size, x.size))  # 初始化一个空向量，用于记录每个时间t对应的指定的x对应的y值
            fv[0, :] = self.functionize(yv)(x)
            for j, f in zip(range(1, len(self.tv)), fv[1:]):
                yv = self.evolve(j, yv, -self.dt)
                # self.v_grid[1:-1, j] = yv  # 记录每个时间t对应的期权价值向量
                f[:] = self.functionize(yv)(x)
            return yv, [LinearFlat(self.tv, f) for f in fv.T]  # 返回边界条件函数bc_bound，input时间t，output边界条件函数值



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
import numpy as np
from pricelib.common.time import global_evaluation_date
from pricelib.common.utilities.enums import TouchType, ExerciseType, PaymentType
from pricelib.common.processes import StochProcessBase
from pricelib.common.pricing_engine_base import AnalyticEngine


class AnalyticDoubleDigitalEngine(AnalyticEngine):
    """双边二元期权闭式级数近似解定价引擎
    Hui(1996) One-Touch Double Barrier Binary Option Values
    只支持美式双接触/双不接触, 到期支付；双接触期权的下边界payoff和上边界payoff必须相等"""

    def __init__(self, stoch_process: StochProcessBase = None, series_num=10, *,
                 s=None, r=None, q=None, vol=None):
        """构造函数
        Args:
            stoch_process: 随机过程StochProcessBase对象
            series_num: int，计算Hui(1996)近似解时级数项的数量
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process=stoch_process, s=s, r=r, q=q, vol=vol)
        self.i_vec = np.array(range(1, series_num + 1))  # 用于计算近似解的级数项

    # pylint: disable=invalid-name, too-many-locals
    def calc_present_value(self, prod, t=None, spot=None):
        assert (prod.payment_type == PaymentType.Expire and
                prod.exercise_type == ExerciseType.American), "Error: Hui(1996)双边二元期权级数近似解, 只支持美式双接触/双不接触, 到期支付"

        calculate_date = global_evaluation_date() if t is None else t
        _maturity = prod.trade_calendar.business_days_between(calculate_date, prod.end_date) / prod.t_step_per_year

        if spot is None:
            spot = self.process.spot()
        r = self.process.interest(_maturity)
        q = self.process.div(_maturity)
        vol = self.process.vol(_maturity, spot)

        if prod.discrete_obs_interval is not None:
            # 均匀离散观察，M. Broadie, P. Glasserman, S.G. Kou(1997) 在连续障碍期权解析解上加调整项，调整障碍价格水平
            # 指数上的beta = -zeta(1/2) / sqrt(2*pi) = 0.5826, 其中zeta是黎曼zeta函数
            bound = [None, None]
            bound[1] = prod.bound[1] * np.exp(0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
            bound[0] = prod.bound[0] * np.exp(-0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
        else:  # 连续观察
            bound = prod.bound

        Z = math.log(bound[1] / bound[0])
        alpha = - 0.5 * (2 * (r - q) / vol ** 2 - 1)
        beta = - 0.25 * (2 * (r - q) / vol ** 2 - 1) ** 2 - 2 * r / vol ** 2

        value0 = self.i_vec * np.pi / Z
        value1 = 2 * np.pi * self.i_vec * prod.rebate[0] / Z ** 2
        value2 = (spot / bound[0]) ** alpha - np.power(-1, self.i_vec) * (spot / bound[1]) ** alpha
        value3 = alpha ** 2 + value0 ** 2
        value4 = np.sin(value0 * np.log(spot / bound[0]))
        value5 = np.exp(-0.5 * vol ** 2 * _maturity * (value0 ** 2 - beta))
        value_series = value1 * value2 * value4 * value5 / value3  # 期权价值
        double_no_touch = np.sum(value_series)

        if prod.touch_type == TouchType.NoTouch:
            return double_no_touch
        if prod.touch_type == TouchType.Touch:
            assert prod.rebate[0] == prod.rebate[1], "Error: Hui(1996)双边二元期权闭式解, 双接触期权的下边界payoff和上边界payoff必须相等"
            return prod.rebate[0] * np.exp(-r * _maturity) - double_no_touch
        raise ValueError("Hui(1996)双边二元期权闭式解, 未知的触碰类型")



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import ExerciseType
from pricelib.common.time import global_evaluation_date
from pricelib.common.pricing_engine_base import BiTreeEngine


class BiTreeVanillaEngine(BiTreeEngine):
    """香草期权二叉树定价引擎
    支持欧式期权和美式期权"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        tau = (prod.end_date - calculate_date).days / prod.annual_days.value
        if spot is None:
            spot = self.process.spot()

        self.r = self.process.interest(tau)
        n_step = round(tau * self.tree_branches)
        s_paths = self.path_generator(n_step, tau, spot).copy()
        v_grid = np.zeros(shape=s_paths.shape)

        for j in range(n_step, -1, -1):
            if j == n_step:
                v_grid[:, j] = np.maximum(prod.callput.value * (s_paths[:, j] - prod.strike), 0)
            else:
                v_grid[:j + 1, j] = (v_grid[:j + 1, j + 1] * self.p + v_grid[1:j + 2, j + 1] * (
                        1 - self.p)) * np.exp(-self.r * self.dt)
                if prod.exercise_type == ExerciseType.American:
                    v_grid[:j + 1, j] = np.maximum(v_grid[:j + 1, j],
                                                   prod.callput.value * (self.s_paths[:j + 1, j] - prod.strike))
        return v_grid[0, 0]



from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='pricelib',
    version='1.1.1',
    packages=find_packages(),
    description='pricelib is an open-source financial derivatives pricing library written in Python.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='上海凌瓴信息科技有限公司',
    author_email='marx@galatech.com.cn',
    url='https://gitee.com/lltech/pricelib',
    install_requires=[
        'numpy>=1.24.1, <=1.26.4',
        'numba>=0.57.1, <=0.59.1',
        'scipy>=1.8.0, <=1.13.0',
        'pandas>=2.0.3, <=2.2.2',
        'rocket-fft~=0.2',
        'importlib-metadata>=6.8.0, <=7.1.0',
    ],
    extras_require={
        'dev': [
            'pytest',
            'flake8',
            'pylint',
        ],
        'plot': [
            'matplotlib>=3.5.3, <=3.7.5',
            'plotly>=5.16.1, <=5.22.0',
        ]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
)



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from contextlib import suppress
from pricelib.common.time import (global_evaluation_date, CN_CALENDAR, AnnualDays)
from pricelib.common.utilities.patterns import Observer
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.product_base.option_base import OptionBase
from pricelib.pricing_engines.mc_engines import MCRangeAccuralEngine


class RangeAccural(OptionBase, Observer):
    """区间累计期权"""

    def __init__(self, s0, upper_strike=None, lower_strike=None, payment=0.1, engine=None, status=0,
                 maturity=None, start_date=None, end_date=None, trade_calendar=CN_CALENDAR,
                 annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        """构造函数
        产品参数:
            s0: float，标的初始价格
            upper_strike: float，高行权价
            lower_strike: float，低行权价
            payment: float，区间内行权收益率，百分比
            engine: 定价引擎，PricingEngine类，仅支持Monte Carlo模拟 MCRangeAccuralEngine
            status: int，之前的累计次数，期初时是0
        时间参数: 要么输入年化期限，要么输入起始日和到期日
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
        可选参数:
            若未提供引擎的情况下，提供了标的价格、无风险利率、分红/融券率、波动率，
            则默认使用蒙特卡洛模拟定价引擎 (BSM模型)
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.s0 = s0  # 标的初始价格
        self.trade_calendar = trade_calendar  # 交易日历
        self.annual_days = annual_days
        self.t_step_per_year = t_step_per_year
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        if maturity is None:
            self.maturity = (end_date - start_date).days / annual_days.value
        else:
            self.maturity = maturity
        self.upper_strike = upper_strike
        self.lower_strike = lower_strike
        self.payment = payment  # 区间内行权收益率，百分比
        self.status = status
        if engine is not None:
            self.set_pricing_engine(engine)
        elif s is not None and r is not None and q is not None and vol is not None:
            self.engine = MCRangeAccuralEngine(s=s, r=r, q=q, vol=vol)

    def set_pricing_engine(self, engine):
        with suppress(AttributeError):
            self.engine.process.spot.remove_observer(self)
        self.engine = engine
        self.engine.process.spot.add_observer(self)
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.engine.process.spot.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """收到标的价格变化的通知时，自动更新累计期权的累计次数"""
        if observable == self.engine.process.spot and self.lower_strike <= observable.value <= self.upper_strike:
            self.status += 1
            logging.info(f"{self}当前累计次数为{self.status}")

    def __repr__(self):
        """返回期权的描述"""
        return "区间累计(Range Accural)"

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .timeutils import *
from .calendars import *

__all__ = ['global_evaluation_date', 'set_evaluation_date', 'get_last_business_date', 'CN_CALENDAR', 'Schedule',
           'BusinessConvention', 'MonthAdjustment', 'EndOfMonthModify', 'AnnualDays']


def get_last_business_date(trade_calendar=CN_CALENDAR):
    """获取估值日期"""
    today = datetime.datetime.now().date()
    if today in trade_calendar.holidays or today.weekday() > 4:
        evaluationDate = trade_calendar.advance(today, datetime.timedelta(days=-1))
    else:
        evaluationDate = today
    return evaluationDate


# 估值日期全局变量
__EvaluationDate = get_last_business_date()


def set_evaluation_date(date):
    """设置估值日期"""
    global __EvaluationDate
    __EvaluationDate = date


def global_evaluation_date():
    """获取估值日期"""
    return __EvaluationDate



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import StatusType
from pricelib.common.time import CN_CALENDAR, AnnualDays
from .autocallable_base import AutocallableBase


class EarlyProfitSnowball(AutocallableBase):
    """早利雪球结构产品类"""

    def __init__(self, s0, barrier_out, barrier_in, coupon_out1, coupon_out2, coupon_div=None, lock_term=3,
                 coupon_stair_ends=np.array([12, 24]), maturity=None, start_date=None, end_date=None, parti_in=1,
                 margin_lvl=1, obs_dates=None, pay_dates=None, status=StatusType.NoTouch, engine=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        """继承自动赎回基类AutocallableBase的参数，详见AutocallableBase的__init__方法
        Args: 以下为区别于经典雪球的参数
            coupon_out1: 第一阶段敲出票息，百分比，年化
            coupon_out2: 第二阶段敲出票息，百分比，年化
            coupon_stair_ends: tuple(int), 分段票息节点, 通常为(12, 24), 代表第几个月末
        """
        super().__init__(s0=s0, maturity=maturity, start_date=start_date, end_date=end_date, lock_term=lock_term,
                         trade_calendar=trade_calendar, obs_dates=obs_dates, pay_dates=pay_dates, status=status,
                         annual_days=annual_days, parti_in=parti_in, margin_lvl=margin_lvl,
                         t_step_per_year=t_step_per_year, engine=engine, s=s, r=r, q=q, vol=vol)
        len_obs_dates = len(self.obs_dates.date_schedule)

        self.barrier_out = np.ones(len_obs_dates) * barrier_out
        self.barrier_in = np.ones(len_obs_dates) * barrier_in
        coupon_out = np.ones(len_obs_dates + lock_term - 1)
        coupon_out[:coupon_stair_ends[0]] = coupon_out1
        coupon_out[coupon_stair_ends[0]:coupon_stair_ends[1]] = coupon_out2
        self.coupon_out = coupon_out[lock_term - 1:]
        self.coupon_div = coupon_div if coupon_div is not None else coupon_out[-1]
        self.parti_out = 0
        self.strike_upper = s0
        self.strike_lower = 0
        self.t_step_per_year = t_step_per_year

    def __repr__(self):
        """返回期权的描述"""
        return "早利雪球"



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""

import math
from scipy.stats import norm
from pricelib.common.utilities.enums import AverageMethod, AsianAveSubstitution, CallPut
from pricelib.common.time import global_evaluation_date
from pricelib.common.pricing_engine_base import AnalyticEngine


class AnalyticAsianEngine(AnalyticEngine):
    """亚式期权闭式解定价引擎  # todo:若定价时已经处于均价观察期内，需要获得当前已经实现的均价，并对行权价进行调整
    Kemma Vorst 1990 支持几何平均价格替代标的资产价格的亚式期权. 仅支持期初估值. 连续观察，平均价观察期为整个期权有效期内
    Turnbull Wakeman 1991 近似估计 支持算术平均价格替代标的资产价格的亚式期权. 平均观察期的起点任意，终点为期权到期日
    todo: 利用固定行权价与浮动行权价亚式期权的对称性，定价平均行权价亚式"""

    # pylint: disable=invalid-name, too-many-locals
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        assert not prod.enhanced, "Error: 亚式期权解析解引擎不支持增强亚式期权"
        assert prod.substitute == AsianAveSubstitution.Underlying, "Error: 亚式期权解析解引擎只支持平均结算价，不支持平均执行价"
        calculate_date = global_evaluation_date() if t is None else t
        # T = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)
        T = _maturity_business_days / prod.t_step_per_year
        obs_start = prod.trade_calendar.business_days_between(calculate_date, prod.obs_start)
        obs_end = prod.trade_calendar.business_days_between(calculate_date, prod.obs_end)
        if spot is None:
            spot = self.process.spot()
        r = self.process.interest(T)
        q = self.process.div(T)
        vol = self.process.vol(T, self.process.spot())
        b = r - q
        if prod.ave_method == AverageMethod.Geometric:
            # Kemna Vorst 1990 几何平均收盘价替代标的资产价格的亚式期权
            assert (obs_start <= 0 and obs_end == _maturity_business_days), \
                "Error: 几何平均亚式期权解析解的观察开始日必须是估值日，观察结束日必须是到期日"
            b_A = 0.5 * (b - vol ** 2 / 6)
            vol_A = vol / math.sqrt(3)

            d1 = (math.log(spot / prod.strike) + (b_A + 0.5 * vol_A ** 2) * T) / (vol_A * math.sqrt(T))
            d2 = d1 - vol_A * math.sqrt(T)

            price = prod.callput.value * (spot * math.exp((b_A - r) * T) * norm.cdf(prod.callput.value * d1)
                                          - prod.strike * math.exp(-r * T) * norm.cdf(prod.callput.value * d2))
            return price

        elif prod.ave_method == AverageMethod.Arithmetic:
            # Turnbull Wakeman 1991近似估计 算术平均收盘价替代标的资产价格的亚式期权
            assert obs_end == _maturity_business_days, "Error: 算术平均亚式期权解析解的观察结束日必须是到期日"

            t1 = max(obs_start, 0) / prod.t_step_per_year
            T2 = (obs_end - obs_start) / prod.t_step_per_year
            tau = T2 - T
            if b == 0:  # 此时公式可以简化
                M1 = 1
            else:
                M1 = (math.exp(b * T) - math.exp(b * t1)) / (b * T - b * t1)
            if tau > 0:  # 若此时在资产价格平均期内，tau = T2 - T > 0
                assert prod.s_average is not None, "Error: 已经处于资产价格平均期内时，必须输入已实现的资产价格平均值s_average"
                strike_hat = T2 / T * prod.strike - tau / T * prod.s_average
                if strike_hat < 0:  # 此时认购期权一定会行权，认沽期权必然不是实值期权，价格为0
                    if prod.callput == CallPut.Call:
                        # 到期时的标的资产平均价格的期望值
                        s_a = prod.s_average * (T2 - T) / T2 + spot * M1 * T / T2
                        return max(s_a - prod.strike, 0) * math.exp(-r * T)
                    else:
                        return 0
            else:
                strike_hat = prod.strike

            if b == 0:
                M2 = (2 * math.exp(vol ** 2 * T) - 2 * math.exp(vol ** 2 * t1) * (1 + vol ** 2 * (T - t1))) / (
                        vol ** 4 * (T - t1) ** 2)
            else:
                M2 = (2 * math.exp((2 * b + vol ** 2) * T)) / ((b + vol ** 2) * (2 * b + vol ** 2) * (T - t1) ** 2) + \
                     (2 * math.exp((2 * b + vol ** 2) * t1)) / (b * (T - t1) ** 2) * (
                             1 / (2 * b + vol ** 2) - (math.exp(b * (T - t1)) / (b + vol ** 2)))

            b_A = math.log(M1) / T
            vol_A = math.sqrt(math.log(M2) / T - 2 * b_A)

            d1 = (math.log(spot / strike_hat) + (b_A + 0.5 * vol_A ** 2) * T) / (vol_A * math.sqrt(T))
            d2 = d1 - vol_A * math.sqrt(T)

            price = prod.callput.value * (spot * math.exp((b_A - r) * T) * norm.cdf(prod.callput.value * d1)
                                          - strike_hat * math.exp(-r * T) * norm.cdf(prod.callput.value * d2))
            if tau > 0:
                price = price * T / T2
            return price

        else:
            raise ValueError("未知的平均方式，亚式平均方式为几何平均/算术平均")



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.pricing_engine_base import McEngine
from pricelib.common.time import global_evaluation_date


class MCPortfolioEngine(McEngine):
    """香草组合 Monte Carlo 模拟定价引擎"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        if spot is None:
            spot = self.process.spot()
        else:
            self.reset_paths_flag()  # 重置路径标志位，重新生成路径
        price = 0
        n_step = round(_maturity * prod.t_step_per_year)
        paths = self.path_generator(n_step=n_step, spot=spot, t_step_per_year=prod.t_step_per_year).copy()
        for vanilla, position in prod.vanilla_list:
            price += np.mean(np.maximum(vanilla.callput.value * (paths[-1, :] - vanilla.strike),
                                        0)) * position * self.process.interest.disc_factor(_maturity)
        return price



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from abc import ABCMeta, abstractmethod
from contextlib import suppress
import datetime
import numpy as np
from ..time import global_evaluation_date
from ..pricing_engine_base.engine_base import PricingEngineBase


class OptionBase(metaclass=ABCMeta):
    """期权产品基类"""

    def __init__(self):
        """初始化
        engine: PricingEngine，定价引擎
        trade_calendar: TradeCalendar，交易日历
        end_date: datetime.date，到期日
        """
        self.engine = None
        self.trade_calendar = None
        self.end_date = None

    @abstractmethod
    def set_pricing_engine(self, engine):
        """设置定价引擎，指定当前的定价函数_pricing_func.
        根据产品类型，可能存在闭式解、蒙特卡洛模拟、PDE有限差分、积分法、树方法等定价方法。
        """

    @abstractmethod
    def __repr__(self):
        """返回期权的描述"""

    @abstractmethod
    def price(self, *args, **kwargs):
        """执行定价
        对于标的期初价格为s0的期权，以香草期权为例，返回的是BSM公式的定价结果，即max(0, St - K)期望的现值
        如果想要得到单位估值，需要再除以s0"""
        self.validate_parameters()

    def validate_parameters(self, t=None):
        """price方法中，先检查产品是否已经配置了定价引擎，估值日是否小于终止日
        Args:
            t: datetime.date，计算期权价格的日期，默认None，此时使用全局估值日"""
        if self.engine is None or not isinstance(self.engine, PricingEngineBase):
            raise AttributeError("尚未配置定价引擎，您可以在以下两种方法中选择:\n"
                                 "    1. 在产品类同时输入s,r,q,vol以创建默认定价引擎;\n"
                                 "    2. 自行创建定价引擎，再使用set_pricing_engine方法配置到产品中")
        calculate_date = global_evaluation_date() if t is None else t
        if calculate_date > self.end_date:
            raise ValueError(f"估值日必须小于等于到期日, 当前估值日为{calculate_date}, 到期日为{self.end_date}.\n"
                             f"默认估值日是今天。如需修改，请使用set_evaluation_date(datetime.date(2024, 5, 20))配置全局估值日。")

    def delta(self, spot=None, step=None, *args, **kwargs):
        """求价格spot处的delta值
        会先尝试调用定价引擎的delta方法，如果没有则使用通用的差分法计算
        Args:
            spot: float，价格的绝对值
            step: float，微扰步长，默认为价格spot的1%
        Returns: delta = ∂V/∂S
        """
        try:
            return self.engine.delta(prod=self, spot=spot, step=step, *args, **kwargs)
        except AttributeError:
            self.validate_parameters()
            spot = self.engine.process.spot() if spot is None else spot
            step = spot * 0.01 if step is None else step
            up_price = self.price(spot=spot + step)
            down_price = self.price(spot=spot - step)
            delta = (up_price - down_price) / (2 * step)
            return delta

    def gamma(self, spot=None, step=None, *args, **kwargs):
        """求价格spot处的gamma值
        会先尝试调用定价引擎的gamma方法，如果没有则使用通用的差分法计算
        Args:
            spot: float，价格的绝对值
            step: float，微扰步长，默认为价格spot的1%
        Returns: gamma = ∂2V/∂S2
        """
        try:
            return self.engine.gamma(prod=self, spot=spot, step=step, *args, **kwargs)
        except AttributeError:
            self.validate_parameters()
            spot = self.engine.process.spot() if spot is None else spot
            step = spot * 0.01 if step is None else step
            up_price = self.price(spot=spot + step)
            down_price = self.price(spot=spot - step)
            mid_price = self.price(spot=spot)
            gamma = (up_price - 2 * mid_price + down_price) / (step ** 2)
            return gamma

    def vega(self, spot=None, step=0.01, *args, **kwargs):
        """计算波动率上升1%的vega todo:目前只支持常数波动率
        会先尝试调用定价引擎的vega方法，如果没有则使用通用的差分法计算
        Args:
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
            step: float，微扰步长，默认为1%
        Returns: vega = ∂V/∂σ
        """
        try:
            return self.engine.vega(prod=self, spot=spot, step=step, *args, **kwargs)
        except AttributeError:
            self.validate_parameters()
            spot = self.engine.process.spot() if spot is None else spot
            last_vol = self.engine.process.vol.volval
            last_price = self.price(spot=spot)
            self.engine.process.vol.volval = last_vol + step
            new_price = self.price(spot=spot)
            self.engine.process.vol.volval = last_vol
            vega = (new_price - last_price) / spot
            return vega

    def theta(self, spot=None, step=1, *args, **kwargs):
        """计算一天的theta值
        会先尝试调用定价引擎的theta方法，如果没有则使用通用的差分法计算
        Args:
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
            step: int，时间步长，默认为1天
        Returns: theta = ∂V/∂t
        """
        try:
            return self.engine.theta(prod=self, spot=spot, step=step, *args, **kwargs)
        except AttributeError:
            self.validate_parameters()
            spot = self.engine.process.spot() if spot is None else spot
            step = datetime.timedelta(days=step)
            origin_price = self.price(spot=spot)
            next_date = self.trade_calendar.advance(global_evaluation_date(), step)
            new_price = self.price(t=next_date, spot=spot)
            theta = (new_price - origin_price) / spot
            return theta

    def rho(self, spot=None, step=0.01, *args, **kwargs):
        """计算无风险利率上升1%的rho todo:目前只支持常数无风险利率
        会先尝试调用定价引擎的rho方法，如果没有则使用通用的差分法计算
        Args:
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
            step: float，微扰步长，默认为1%
        Returns: rho = ∂V/∂r
        """
        try:
            return self.engine.rho(prod=self, spot=spot, step=step, *args, **kwargs)
        except AttributeError:
            self.validate_parameters()
            spot = self.engine.process.spot() if spot is None else spot
            last_r = self.engine.process.interest.data
            last_price = self.price(spot=spot)
            self.engine.process.interest.data = last_r + step
            new_price = self.price(spot=spot)
            self.engine.process.interest.data = last_r
            rho = (new_price - last_price) / spot
            return rho

    def pv_and_greeks(self, spot=None, *args, **kwargs):
        """当一次性计算pv和5个greeks时，可以调用此函数，减少计算量
        会先尝试调用定价引擎的pv_and_greeks方法，如果没有则使用通用的差分法计算
        Args:
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
        Returns:
            Dict[str:float]: {'pv': pv, 'delta': delta, 'gamma': gamma, 'theta': theta, 'vega': vega, 'rho': rho}
        """
        if hasattr(self.engine, "pv_and_greeks"):
            with suppress(AttributeError):
                self.price()
                return self.engine.pv_and_greeks(prod=self, spot=spot, *args, **kwargs)
        self.validate_parameters()
        spot = self.engine.process.spot() if spot is None else spot
        pv = np.float64(self.price(spot=spot))
        s_step = spot * 0.01
        up_price = self.price(spot=spot + s_step)
        down_price = self.price(spot=spot - s_step)
        delta = (up_price - down_price) / (2 * s_step)
        gamma = (up_price - 2 * pv + down_price) / (s_step ** 2)

        last_vol = self.engine.process.vol.volval
        self.engine.process.vol.volval = last_vol + 0.01
        new_price = self.price(spot=spot)
        self.engine.process.vol.volval = last_vol
        vega = (new_price - pv) / spot

        last_r = self.engine.process.interest.data
        self.engine.process.interest.data = last_r + 0.01
        new_price = self.price(spot=spot)
        self.engine.process.interest.data = last_r
        rho = (new_price - pv) / spot

        t_step = datetime.timedelta(days=1)
        next_date = self.trade_calendar.advance(global_evaluation_date(), t_step)
        new_price = self.price(t=next_date, spot=spot)
        theta = (new_price - pv) / spot
        return {'pv': pv, 'delta': delta, 'gamma': gamma, 'vega': vega, 'theta': theta, 'rho': rho}



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from contextlib import suppress
from pricelib.common.utilities.enums import TouchType, EngineType, ExerciseType, PaymentType, StatusType
from pricelib.common.time import CN_CALENDAR, AnnualDays, global_evaluation_date
from pricelib.common.utilities.patterns import Observer
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.product_base.option_base import OptionBase
from pricelib.pricing_engines.analytic_engines import AnalyticDoubleDigitalEngine
from pricelib.pricing_engines.fdm_engines import FdmDigitalEngine


class DoubleDigitalOption(OptionBase, Observer):
    """双边二元期权
        欧式: 二元凹式/二元凸式
        美式: 美式双接触/美式双不接触"""

    def __init__(self, touch_type: TouchType, exercise_type: ExerciseType, payment_type: PaymentType,
                 *, bound=(80, 120), rebate=(0, 0), engine=None, status=StatusType.NoTouch,
                 maturity=None, start_date=None, end_date=None, discrete_obs_interval=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        """构造函数
        产品参数:
            touch_type: 接触方式 - 接触Touch/不接触NotTouch
            exercise_type: 行权方式，ExerciseType枚举类，European/American
            payment_type: 支付方式，PaymentType枚举类，立即支付Hit/到期支付Expire
            bound: (float, float), (低障碍价, 高障碍价)
            rebate: (float, float), 触碰payoff的绝对数值，非年化
                    双接触: (下边界payoff, 上边界payoff); 双不接触: (到期不接触的payoff, 无作用)
            status: 定价时，触碰障碍的状态，StatusType枚举类，默认为NoTouch未触碰，UpTouch为向上触碰，DownTouch为向下触碰
            engine: 定价引擎，PricingEngine类
                    解析解: AnalyticDoubleDigitalEngine - Hui(1996) 级数近似解，双接触期权的下边界payoff和上边界payoff必须相等
                                                        支持 连续/离散观察、到期支付的(美式)双接触/双不接触
                    PDE: FdmDigitalEngine 支持 (欧式)二元凹式/二元凸式；支持 连续/离散观察、立即/到期支付的(美式)双接触/美式双不接触
                    蒙特卡洛: MCDoubleDigitalEngine 支持 (欧式)二元凹式/二元凸式；只支持离散观察、立即/到期支付的(美式)双接触/美式双不接触
        时间参数: 要么输入年化期限，要么输入起始日和到期日
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
            discrete_obs_interval: 观察时间间隔. 若为连续观察，None；若为均匀离散观察，为年化的观察时间间隔
        可选参数:
            若未提供引擎的情况下，提供了标的价格、无风险利率、分红/融券率、波动率，
            则使用默认定价引擎：
                到期支付的美式双不接触默认使用 Hui(1996) 级数近似解 AnalyticDoubleDigitalEngine
                其余情况默认使用 PDE 定价引擎 FdmDigitalEngine
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.trade_calendar = trade_calendar
        self.annual_days = annual_days
        self.t_step_per_year = t_step_per_year
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        self.maturity = maturity if maturity is not None else (self.end_date - self.start_date).days / annual_days.value
        self.bound = bound  # (低障碍价, 高障碍价)
        self.rebate = rebate  # 双接触: (下边界payoff, 上边界payoff); 双不接触: (不接触payoff, 无作用)
        self.touch_type = touch_type
        self.exercise_type = exercise_type
        self.payment_type = payment_type
        self.discrete_obs_interval = discrete_obs_interval  # 连续观察=None；均匀离散观察=观察时间间隔
        self.status = status
        if engine is not None:
            self.set_pricing_engine(engine)
        elif s is not None and r is not None and q is not None and vol is not None:
            if (exercise_type == ExerciseType.American and payment_type == PaymentType.Expire
                    and touch_type == TouchType.NoTouch):
                default_engine = AnalyticDoubleDigitalEngine(s=s, r=r, q=q, vol=vol)
                self.set_pricing_engine(default_engine)
            else:
                default_engine = FdmDigitalEngine(s=s, r=r, q=q, vol=vol)
                self.set_pricing_engine(default_engine)

    def set_pricing_engine(self, engine):
        """设置定价引擎，同时将自己注册为观察者。若已有定价引擎，先将自己从原定价引擎的观察者列表中移除"""
        with suppress(AttributeError, ValueError):
            self.engine.process.spot.remove_observer(self)
        self.engine = engine
        self.engine.process.spot.add_observer(self)
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.engine.process.spot.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """收到标的价格变化的通知时，自动更新是否已经触碰障碍"""
        if observable == self.engine.process.spot:
            if self.discrete_obs_interval is None:  # TODO: 目前update仅支持连续观察，离散观察需要判断是否是观察日
                if observable.value >= self.bound[1]:
                    self.status = StatusType.UpTouch
                elif observable.value <= self.bound[0]:
                    self.status = StatusType.DownTouch

    def __repr__(self):
        """返回期权的描述"""
        if self.exercise_type == ExerciseType.European:
            if self.touch_type == TouchType.Touch:
                return f"{self.payment_type.value}二元凹式(欧式双接触期权)"
            else:  # NoTouch
                return f"{self.payment_type.value}二元凸式(欧式双不接触期权)"
        elif self.exercise_type == ExerciseType.American:
            if self.discrete_obs_interval is None:
                obs_type = "连续观察"
            else:
                obs_type = "离散观察"
            return f"{self.payment_type.value}{obs_type}美式双{self.touch_type.value}二元期权"
        else:
            raise ValueError("无法识别的期权类型")

    @time_this
    def price(self, t=None, spot=None):
        self.validate_parameters(t=t)
        if self.engine.engine_type == EngineType.PdeEngine:  # 接口特殊是因为PDE引擎兼容了单双边障碍
            price = self.engine.calc_present_value(prod=self, t=t, spot=spot, bound=self.bound, rebate=self.rebate)
        else:
            price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from abc import ABCMeta, abstractmethod
import datetime
from ..utilities.enums import EngineType
from ..utilities.patterns import SimpleQuote
from ..utilities.utility import logging
from ..term_structures import ConstantRate
from ..volmodels import BlackConstVol
from ..processes import StochProcessBase, GeneralizedBSMProcess


class PricingEngineBase(metaclass=ABCMeta):
    """定价引擎基类"""
    engine_type: EngineType = None  # 定价引擎类型，枚举类

    def __init__(self, stoch_process=None, s=None, r=None, q=None, vol=None):
        """构造函数，若stoch_process不为None，优先设置随机过程为stoch_process，忽略s、r、q、vol参数；
                   若stoch_process=None，默认构造BSM随机过程，需要输入标的价格、无风险利率、分红/融券率、波动率；
                   若未设置随机过程，也未提供构造随机过程的参数，也可以后续通过set_stoch_process方法设置随机过程
        Args:
            stoch_process: 随机过程StochProcessBase对象
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        if stoch_process is not None:
            self.set_stoch_process(stoch_process)
        elif s is not None and r is not None and q is not None and vol is not None:
            spot_price = SimpleQuote(value=s, name="标的价格")
            riskfree = ConstantRate(value=r, name="无风险利率")
            dividend = ConstantRate(value=q, name="分红/融券率")
            volatility = BlackConstVol(vol, name="波动率")
            self.process = GeneralizedBSMProcess(spot=spot_price, interest=riskfree, div=dividend, vol=volatility)
        elif not all(var is None for var in [s, r, q, vol]) and any(var is None for var in [s, r, q, vol]):
            raise ValueError("构造定价引擎时，当前价格s、无风险利率r、分红/融券率q、波动率vol必须同时输入，才会创建默认定价引擎")
        else:
            self.process = None
            logging.warning(
                "构造定价引擎时，未设置随机过程，也未提供构造随机过程的参数，无法初始化随机过程。请调用set_stoch_process方法设置随机过程!")

    def set_stoch_process(self, stoch_process: StochProcessBase):
        """设置随机过程，其中包含了价格动态过程及波动率信息
        Args:
            stoch_process: 随机过程StochProcessBase对象
        """
        self.process = stoch_process

    @abstractmethod
    def calc_present_value(self, prod, t: datetime.date = None, spot: float = None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .fdm_vanilla_engine import *
from .fdm_barrier_engine import *
from .fdm_digital_engine import *
from .fdm_double_shark_engine import *
from .fdm_airbag_engine import *
from .fdm_autocallable_engine import *

__all__ = ['FdmVanillaEngine', 'FdmBarrierEngine', 'FdmDigitalEngine', 'FdmDoubleSharkEngine', 'FdmAirbagEngine',
           'FdmSnowBallEngine', 'FdmPhoenixEngine', ]



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from pricelib.common.utilities.enums import PaymentType, InOut, CallPut, ExerciseType, TouchType
from pricelib.common.processes import StochProcessBase
from pricelib.common.pricing_engine_base import AnalyticEngine
from ...products.barrier.double_barrier_option import DoubleBarrierOption
from ...products.digital.double_digital_option import DoubleDigitalOption
from .analytic_double_barrier_engine import AnalyticDoubleBarrierEngine
from .analytic_double_digital_engine import AnalyticDoubleDigitalEngine


class AnalyticDoubleSharkEngine(AnalyticEngine):
    """双鲨结构闭式解定价引擎, 由双边敲出看涨与双边敲出看跌组合而成，现金返还使用美式双接触闭式解
        Haug(1998) Put-Call Barrier Transformations 对于标的持有成本为0的期权(期货期权)，利用认购-认沽障碍期权对称性，
                                                    利用单边障碍期权解析解的组合给双障碍期权定价。
        Ikeda and Kunitomo(1992)将双障碍期权表示为加权正态分布函数的无限集，但是仅当行权价在障碍范围内时，公式才是成立的。
    Haug(1998)的解析解与Ikeda and Kunitomo(1992)的结果几乎一致，两者对比，
        Ikeda and Kunitomo(1992)适用于标的持有成本不为0的情形, Haug(1998)适用于行权价在障碍范围之外的情形。
    由于现金返还使用了美式双接触闭式解Hui(1996)，双接触期权的下边界payoff和上边界payoff必须相等，
        所以双鲨闭式解的高低敲出现金返还也必须相等，现金返还到期支付
    只支持美式观察(整个有效期观察)；支持连续观察/离散观察(默认为每日观察)
    """

    def __init__(self, stoch_process: StochProcessBase = None, formula_type="Ikeda&Kunitomo1992",
                 series_num=10, delta1=0, delta2=0, *,
                 s=None, r=None, q=None, vol=None):
        """
        Args:
            stoch_process: 随机过程StochProcessBase对象
            self.formula_type: 解析解类型，Ikeda&Kunitomo1992或Haug1998
        其他三个参数是Ikeda&Kunitomo1992的参数：
            self.series_num: 用于计算近似解的级数项 默认 i = -10 ~ 10
            self.delta1: 上边界的曲率
            self.delta2: 下边界的曲率，
                当delta1 = delta2 = 0时，对应两条平行边界
                当delta1 < 0 < delta2 时，对应下边界随时间呈指数增长，而上百年姐随时间呈指数衰减
                当delta1 > 0 > delta2 时，对应一个向下凸的下边界和一个向上凸的上边界
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process=stoch_process, s=s, r=r, q=q, vol=vol)
        self.formula_type = formula_type
        self.series_num = series_num
        self.delta1 = delta1
        self.delta2 = delta2

    # pylint: disable=invalid-name
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        assert (prod.exercise_type == ExerciseType.American and prod.payment_type == PaymentType.Expire), \
            "Error: 双鲨解析解仅支持美式观察、到期支付的情形。"
        assert (prod.strike[0] <= prod.strike[1]) and (
                prod.bound[0] < prod.bound[1]), "Error: 双鲨结构输入的行权价和障碍价必须是递增的，不能是None"

        DOC = DoubleBarrierOption(exercise_type=ExerciseType.American, payment_type=prod.payment_type, inout=InOut.Out,
                                  callput=CallPut.Call, strike=prod.strike[1], bound=prod.bound, rebate=(0, 0),
                                  window=prod.window, start_date=prod.start_date, end_date=prod.end_date,
                                  discrete_obs_interval=prod.discrete_obs_interval,
                                  engine=AnalyticDoubleBarrierEngine(self.process, formula_type=self.formula_type,
                                                                     series_num=self.series_num, delta1=self.delta1,
                                                                     delta2=self.delta2))
        DOP = DoubleBarrierOption(exercise_type=ExerciseType.American, payment_type=prod.payment_type, inout=InOut.Out,
                                  callput=CallPut.Put, strike=prod.strike[0], bound=prod.bound, rebate=(0, 0),
                                  window=prod.window, start_date=prod.start_date, end_date=prod.end_date,
                                  discrete_obs_interval=prod.discrete_obs_interval,
                                  engine=AnalyticDoubleBarrierEngine(self.process, formula_type=self.formula_type,
                                                                     series_num=self.series_num, delta1=self.delta1,
                                                                     delta2=self.delta2))
        double_touch = DoubleDigitalOption(bound=prod.bound, rebate=prod.rebate, exercise_type=ExerciseType.American,
                                           payment_type=prod.payment_type, touch_type=TouchType.Touch,
                                           start_date=prod.start_date, end_date=prod.end_date,
                                           discrete_obs_interval=prod.discrete_obs_interval,
                                           engine=AnalyticDoubleDigitalEngine(self.process, series_num=self.series_num))
        result = (DOC.price(t=t, spot=spot) * prod.parti[1] +
                  DOP.price(t=t, spot=spot) * prod.parti[0] +
                  double_touch.price(t=t, spot=spot))
        return result



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from .fdm_barrier_engine import FdmBarrierEngine


class FdmAirbagEngine(FdmBarrierEngine):
    """安全气囊PDE有限差分法定价引擎
    安全气囊是向下敲入看涨期权，到期支付
    支持连续观察/离散观察(默认为每日观察)"""

    def _init_boundary_condition(self, smax, maturity):
        """初始化边界条件
        安全气囊，未敲入是call，已敲入是标的资产
        Args:
            smax: float, 有限差分网格上边界的价格最大值
            maturity: float, 到期时间
        Returns:
            self.fn_callput: List[Callable], 0~smax网格的边界条件
            self.fn_bound: List[Callable], 障碍价格网格的边界条件
        """
        # 生成无障碍价格的边界条件： 0~smax
        self.fn_callput = [lambda u: -self.prod.strike * self.prod.knockin_parti *
                                     self.process.interest.disc_factor(maturity, u),
                           lambda u: (smax * self.process.div.disc_factor(maturity, u)
                                      - self.prod.strike * self.process.interest.disc_factor(maturity, u)
                                      ) * self.prod.knockin_parti]
        # 生成障碍价格的Dirichlet边界条件： bound[0]~smax,  第u天触碰障碍产生的payoff，折现到第u天的价值。
        # 返回一个包含两个lambda函数的列表，列表索引分别为[0, 1]，每个lambda函数的形参r=self.rebate中的r补偿值
        self.fn_bound = []
        for i, r in enumerate(self.rebate):
            if self.bound[i] in [-np.inf, np.inf]:  # 安全气囊未敲入call的上边界是smax, 没有上方的障碍，即障碍价无穷
                self.fn_bound.append(lambda u: (smax * self.process.div.disc_factor(maturity, u) -
                                                self.prod.strike * self.process.interest.disc_factor(maturity, u))
                                               * self.prod.call_parti)
            else:
                # 'PaymentMethod.Expire' 触碰下方的障碍后，到期时支付，安全气囊已敲入是标的资产
                if self.prod.discrete_obs_interval is None:  # 连续观察, 价格下边界是self.bound[i]
                    self.fn_bound.append(lambda u: (self.bound[i] - self.prod.strike) * self.prod.knockin_parti
                                                   * self.process.interest.disc_factor(maturity, u))
                else:  # 离散观察，价格下边界是0
                    self.fn_bound.append(
                        lambda u: - self.prod.strike * self.prod.knockin_parti * self.process.interest.disc_factor(
                            maturity, u))

    def _init_terminal_condition(self, s_vec):
        """初始化终止条件
        连续观察（网格上下界是障碍价格）：初始化终止时已敲入的期权价值，安全气囊敲入后是标的资产
        离散观察（网格上下界是0和smax）：初始化终止时的期权价值
        """
        yv = np.where(s_vec - self.prod.strike > 0,
                      (s_vec - self.prod.strike) * self.prod.reset_call_parti,  # 敲入的重置后看涨参与率
                      (s_vec - self.prod.strike) * self.prod.knockin_parti)  # 敲入的下跌参与率
        return yv

    def _init_no_touch_terminal_condition(self, s_vec, p=0, maturity=0):
        """初始化障碍观察区间终点未敲入的期权价值，安全气囊，未敲入是call"""
        if self.prod.discrete_obs_interval is None:  # 连续观察
            yv = np.maximum(s_vec - self.prod.strike, 0) * self.process.interest.disc_factor(maturity,  # 未敲入的看涨参与率
                                                                                             p) * self.prod.call_parti
        else:  # 离散观察
            yv = np.where(s_vec - self.bound[0] > 0,
                          np.maximum(s_vec - self.prod.strike, 0) * self.prod.call_parti,  # 未敲入的看涨参与率
                          (s_vec - self.prod.strike) * self.prod.knockin_parti)  # 敲入的下跌参与率
        return yv



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from contextlib import suppress
from pricelib.common.utilities.enums import CallPut, TouchType, ExerciseType, PaymentType, EngineType, StatusType
from pricelib.common.time import CN_CALENDAR, AnnualDays, global_evaluation_date
from pricelib.common.utilities.patterns import Observer
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.product_base.option_base import OptionBase
from pricelib.pricing_engines.analytic_engines import AnalyticCashOrNothingEngine


class DigitalOption(OptionBase, Observer):
    """单边二元(数字)期权-现金或无(cash or nothing)
    包括欧式二元(现金或无)、美式二元(立即支付)、一触即付(到期支付)，支持M. Broadie, P. Glasserman, S.G. Kou(1999)离散观察调整"""
    touch_type = TouchType.Touch  # 触碰类型：触碰

    def __init__(self, strike, rebate, callput: CallPut, exercise_type: ExerciseType, payment_type: PaymentType, *,
                 maturity=None, start_date=None, end_date=None, engine=None, status=StatusType.NoTouch,
                 discrete_obs_interval=None, trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365,
                 t_step_per_year=243, s=None, r=None, q=None, vol=None):
        """构造函数
        产品参数:
            strike:  float, 执行价
            rebate:  float, 行权收益, 绝对数值（非年化）
            callput: 看涨看跌类型，CallPut枚举类，Call/Put
            exercise_type: 行权方式，ExerciseType枚举类，European/American
            payment_type: 支付方式，PaymentType枚举类，立即支付Hit/到期支付Expire
            status: 触碰障碍的状态，StatusType枚举类，默认为NoTouch未触碰，UpTouch为向上触碰，DownTouch为向下触碰
            engine: 定价引擎，PricingEngine类，以下几种引擎均支持欧式二元(现金或无)、美式二元(立即支付)、一触即付(到期支付)，
                            解析解和PDE引擎支持离散观察/连续观察，其余引擎只支持离散观察(默认为每日观察)
                    解析解: AnalyticCashOrNothingEngine
                    蒙特卡洛: MCDigitalEngine
                    PDE: FdmDigitalEngine
                    积分法: QuadDigitalEngine
                    二叉树: BiTreeDigitalEngine
        时间参数: 要么输入年化期限，要么输入起始日和到期日
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
            discrete_obs_interval: 观察时间间隔. 若为连续观察，None；若为均匀离散观察，为年化的观察时间间隔
        可选参数:
            若未提供引擎的情况下，提供了标的价格、无风险利率、分红/融券率、波动率，
            则默认使用 二元(数字)期权-现金或无(cash or nothing)-闭式解定价引擎 Reiner and Rubinstein(1991b)
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.trade_calendar = trade_calendar
        self.annual_days = annual_days
        self.t_step_per_year = t_step_per_year
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        self.maturity = maturity if maturity is not None else (self.end_date - self.start_date).days / annual_days.value
        self.strike = strike
        self.rebate = rebate  # 绝对值，行权收益
        self.callput = callput
        self.exercise_type = exercise_type
        self.payment_type = payment_type
        self.discrete_obs_interval = discrete_obs_interval  # 连续观察=None；均匀离散观察=观察时间间隔
        self.status = status
        if engine is not None:
            self.set_pricing_engine(engine)
        elif s is not None and r is not None and q is not None and vol is not None:
            default_engine = AnalyticCashOrNothingEngine(s=s, r=r, q=q, vol=vol)
            self.set_pricing_engine(default_engine)

    def set_pricing_engine(self, engine):
        """设置定价引擎，同时将自己注册为观察者。若已有定价引擎，先将自己从原定价引擎的观察者列表中移除"""
        with suppress(AttributeError, ValueError):
            self.engine.process.spot.remove_observer(self)
        self.engine = engine
        self.engine.process.spot.add_observer(self)
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.engine.process.spot.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """收到标的价格变化的通知时，自动更新是否已经触碰障碍"""
        if observable == self.engine.process.spot:
            if self.discrete_obs_interval is None:  # TODO: 目前update仅支持连续观察，离散观察需要判断是否是观察日
                if observable.value >= self.strike and self.callput == CallPut.Call:
                    self.status = StatusType.UpTouch
                elif observable.value <= self.strike and self.callput == CallPut.Put:
                    self.status = StatusType.DownTouch

    def __repr__(self):
        """返回期权的描述"""
        if self.discrete_obs_interval is None:
            obs_type = "连续观察"
        else:
            obs_type = "离散观察"
        return f"{obs_type}{self.exercise_type.value}{self.callput.name}二元期权"

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        if self.engine.engine_type == EngineType.PdeEngine:  # 接口特殊是因为PDE引擎兼容了单双边二元
            if self.callput == CallPut.Call:
                bound = (None, self.strike)
            else:  # Put
                bound = (self.strike, None)
            price = self.engine.calc_present_value(prod=self, t=t, spot=spot, bound=bound,
                                                   rebate=(self.rebate, self.rebate))
        else:
            price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.time import global_evaluation_date
from pricelib.common.utilities.enums import InOut, UpDown
from pricelib.common.pricing_engine_base import McEngine


class MCBarrierEngine(McEngine):
    """障碍期权 Monte Carlo 模拟定价引擎
    只支持离散观察(默认为每日观察)；敲入现金返还为到期支付；敲出现金返还为到期支付"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)
        if spot is None:
            spot = self.process.spot()
        else:
            self.reset_paths_flag()  # 重置路径标志位，重新生成路径
        r = self.process.interest(_maturity)
        paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                    t_step_per_year=prod.t_step_per_year).copy()

        if prod.discrete_obs_interval is None:  # 每日观察
            if prod.updown == UpDown.Up:
                knock_inout = np.any(paths >= prod.barrier, axis=0)
            elif prod.updown == UpDown.Down:
                knock_inout = np.any(paths <= prod.barrier, axis=0)
            else:
                raise ValueError("不支持的UpDown类型")
        else:  # 均匀离散观察
            dt_step = prod.discrete_obs_interval * prod.t_step_per_year
            obs_points = np.flip(np.round(np.arange(_maturity_business_days, 0, -dt_step)).astype(int))
            obs_points = np.concatenate((np.array([0]), obs_points))
            obs_points[obs_points > _maturity_business_days] = _maturity_business_days  # 防止闰年导致的下标越界
            if prod.updown == UpDown.Up:
                knock_inout = np.any(paths[obs_points] >= prod.barrier, axis=0)
            elif prod.updown == UpDown.Down:
                knock_inout = np.any(paths[obs_points] <= prod.barrier, axis=0)
            else:
                raise ValueError("不支持的UpDown类型")

        if prod.inout == InOut.In:
            payoff = np.ones(paths[-1].size) * prod.rebate
            payoff[knock_inout] = np.maximum(prod.callput.value * (paths[-1, knock_inout] - prod.strike),
                                             0) * prod.parti
            price = np.mean(payoff) * np.exp(- r * _maturity)
            return price
        elif prod.inout == InOut.Out:
            # todo: 敲出期权的现金补偿目前是到期支付，添加敲出时立即支付
            payoff = np.maximum(prod.callput.value * (paths[-1] - prod.strike), 0) * prod.parti
            payoff[knock_inout] = prod.rebate
            price = np.mean(payoff) * np.exp(- r * _maturity)
            return price
        else:
            raise ValueError("不支持的InOut类型")



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from functools import lru_cache
import numpy as np
from pricelib.common.utilities.enums import InOut
from pricelib.common.time import global_evaluation_date
from pricelib.common.pricing_engine_base import QuadEngine


class QuadDoubleSharkEngine(QuadEngine):
    """双鲨结构PDE有限差分法定价引擎
    只支持美式观察(整个有效期观察)；只支持离散观察(默认为每日观察)；支持现金返还到期支付"""
    inout = InOut.Out  # 双鲨结构属于敲出期权

    @lru_cache(maxsize=1)
    def set_quad_price_range(self, s_sliced):
        """
        设置数值积分的区间(价格范围)
        Args:
            s_sliced: np.ndarray = self.s[:, step], 其中step是从后向前的步数
        Returns:
            quad_range: tuple, (np.ndarray, ), 数值积分的区间(索引)
            upper_an_range: tuple, (np.ndarray, ), 解析式计算积分的区间(索引)
            lower_an_range: tuple, (np.ndarray, ), 解析式计算积分的区间(索引)
        """
        quad_range = np.where((s_sliced < self.prod.bound[1]) & (s_sliced > self.prod.bound[0]))
        upper_an_range = np.where(s_sliced >= self.prod.bound[1])
        lower_an_range = np.where(s_sliced <= self.prod.bound[0])
        return quad_range, upper_an_range, lower_an_range

    # pylint: disable=too-many-locals
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        self.prod = prod
        calculate_date = global_evaluation_date() if t is None else t
        tau = prod.trade_calendar.business_days_between(calculate_date, prod.end_date) / prod.t_step_per_year
        if spot is None:
            spot = self.process.spot()
        r = self.process.interest(tau)
        q = self.process.div(tau)
        vol = self.process.vol(tau, spot)
        dt = prod.discrete_obs_interval if prod.discrete_obs_interval is not None else 1 / prod.t_step_per_year
        self.backward_steps = int(tau / dt)
        self._check_method_params()

        # v_grid = np.empty(shape=(self.n_points, self.backward_steps + 1))
        # 设置积分法engine参数
        self.set_quad_params(r=r, q=q, vol=vol)
        # 初始化fft的对数价格向量及边界的对数价格向量
        self.init_grid(spot, vol, tau)
        s_vec = np.exp(self.ln_s_vec)
        # 从后向前进行积分计算
        # 设定期末价值，在障碍价格内侧，默认设定未敲入，未敲出
        call_put_payoff = (np.maximum(s_vec - prod.strike[1], 0) * prod.parti[1] +
                           np.maximum(prod.strike[0] - s_vec, 0) * prod.parti[0])
        v_vec = np.where(s_vec >= prod.bound[1], prod.rebate[1],
                         np.where(s_vec <= prod.bound[0], prod.rebate[0], call_put_payoff))

        for step in range(self.backward_steps - 1, 0, -1):
            quad_range, upper_an_range, lower_an_range = self.set_quad_price_range(s_vec)
            # 数值积分计算区域
            y = self.ln_s_vec
            x = self.ln_s_vec[quad_range]
            v_vec[quad_range] = self.fft_step_backward(x, y, v_vec, dt)
            # 解析公式计算积分区域
            v_vec[upper_an_range] = prod.rebate[1] * np.exp(-r * dt * (self.backward_steps - step))
            v_vec[lower_an_range] = prod.rebate[0] * np.exp(-r * dt * (self.backward_steps - step))

        x = np.log(np.array([spot]))
        value = self.fft_step_backward(x, self.ln_s_vec, v_vec, dt)[0]
        return value



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from contextlib import suppress
from pricelib.common.utilities.enums import CallPut, InOut, UpDown, ExerciseType, EngineType, PaymentType, StatusType
from pricelib.common.time import CN_CALENDAR, AnnualDays, global_evaluation_date
from pricelib.common.utilities.patterns import Observer
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.product_base.option_base import OptionBase
from pricelib.pricing_engines.fdm_engines import FdmAirbagEngine


class Airbag(OptionBase, Observer):
    """安全气囊
    安全气囊是美式观察向下敲入看涨期权，到期支付
        敲入前是call，执行价等于合约设立时的标的期初价格
        敲入后payoff与持有标的资产相同(在上涨和下跌参与率为1的情况下)"""
    updown = UpDown.Down
    inout = InOut.In
    callput = CallPut.Call
    payment_type = PaymentType.Expire
    exercise_type = ExerciseType.American

    def __init__(self, strike, barrier, *, knockin_parti=1, call_parti=0.8, reset_call_parti=1, engine=None,
                 status=StatusType.NoTouch, maturity=None, start_date=None, end_date=None, discrete_obs_interval=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        """构造函数
        产品参数:
            strike:  float, 执行价
            barrier:  float, 敲入障碍价格
            knockin_parti: float, 敲入的下跌参与率
            call_parti: float, 未敲入的看涨参与率
            reset_call_parti: float, 敲入的重置后看涨参与率
            status: 敲入状态，StatusType枚举类，默认为NoTouch未敲入未敲出，DownTouch为已敲入
            engine: 定价引擎，PricingEngine类
                    解析解: AnalyticAirbagEngine 由障碍与二元解析解(资产或无)组合而成，所以敲入的下跌参与率与重置后看涨参与率必须相等。
                                                支持连续观察/离散观察(默认为每日观察)
                    PDE: FdmAirbagEngine 支持连续观察/离散观察(默认为每日观察)
                    蒙特卡洛: MCAirbagEngine 只支持离散观察(默认为每日观察)
        时间参数: 要么输入年化期限，要么输入起始日和到期日
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
            discrete_obs_interval: 观察时间间隔. 若为连续观察，None；若为均匀离散观察，为年化的观察时间间隔
        可选参数:
            若未提供引擎的情况下，提供了标的价格、无风险利率、分红/融券率、波动率，
            则默认使用 PDE 定价引擎 FdmAirbagEngine
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.trade_calendar = trade_calendar
        self.annual_days = annual_days
        self.t_step_per_year = t_step_per_year
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        self.maturity = maturity if maturity is not None else (self.end_date - self.start_date).days / annual_days.value
        self.strike = strike
        self.barrier = barrier
        self.knockin_parti = knockin_parti
        self.call_parti = call_parti
        self.reset_call_parti = reset_call_parti
        self.discrete_obs_interval = discrete_obs_interval  # 连续观察=None；均匀离散观察=观察时间间隔
        self.status = status
        if engine is not None:
            self.set_pricing_engine(engine)
        elif s is not None and r is not None and q is not None and vol is not None:
            default_engine = FdmAirbagEngine(s=s, r=r, q=q, vol=vol)
            self.set_pricing_engine(default_engine)

    def set_pricing_engine(self, engine):
        """设置定价引擎，同时将自己注册为观察者。若已有定价引擎，先将自己从原定价引擎的观察者列表中移除"""
        with suppress(AttributeError, ValueError):
            self.engine.process.spot.remove_observer(self)
        self.engine = engine
        self.engine.process.spot.add_observer(self)
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.engine.process.spot.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """收到标的价格变化的通知时，自动更新是否已经触碰障碍，默认每日观察"""
        if observable == self.engine.process.spot:
            if observable.value <= self.barrier:
                self.status = StatusType.DownTouch

    def __repr__(self):
        """返回期权的描述"""
        return "安全气囊"

    @time_this
    def price(self, t=None, spot=None):
        self.validate_parameters(t=t)
        if self.engine.engine_type == EngineType.PdeEngine:  # 接口特殊是因为PDE引擎兼容了单双边障碍
            price = self.engine.calc_present_value(prod=self, t=t, spot=spot, bound=(self.barrier, None),
                                                   rebate=(0, 0))
        else:
            price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .quad_vanilla_engine import *
from .quad_barrier_engine import *
from .quad_digital_engine import *
from .quad_portfolio_engine import *
from .quad_autocall_engine import *
from .quad_snowball_engine import *
from .quad_fcn_engine import *
from .quad_double_shark_engine import *

__all__ = ['QuadVanillaEngine', 'QuadBarrierEngine', 'QuadDigitalEngine', 'QuadPortfolioEngine',
           'QuadAutoCallEngine', 'QuadSnowballEngine', 'QuadFCNEngine', 'QuadDoubleSharkEngine']
